# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Sequence, Union

import torch
import torch.nn as nn
import torch.nn.functional as F

from mmflow.registry import MODELS


def binary_cross_entropy(pred: torch.Tensor, target: torch.Tensor,
                         balance: bool, reduction: str) -> torch.Tensor:
    r"""Calculate (weighted) binary cross entropy between occlusion prediction
    and occlusion ground truth.

    The loss function is
    .. math::
      loss = w \cdot o_pred\ln{o_gt}+\bar{w}(1-o_pred)\ln{(1-o_gt)}

    where the weights is:
    .. math::
      w=\frac{H \cdot W}{\sum{o_pred}+\sum{o_gt}}

    and
    .. math::
      \bar{w}=\frac{H \cdot W}{\sum{(1-o_pred)}+\sum{(1-o_gt)}}

    but from IRR released code, the weights were scaled by 0.5.

    Args:
        pred (Tensor): output predicted occ map from flow_estimator
            with shape(B, 1, H, W).
        target (Tensor): ground truth occ map with shape (B, 1, H, W).
        balance (bool): whether balance the class weights for irr models.
        reduction (str): the reduction to apply to the output:'none', 'mean',
            'sum'. 'none': no reduction will be applied and will return a
            full-size epe map, 'mean': the mean of the epe map is taken, 'sum':
            the epe map will be summed but averaged by batch_size.
            Default: 'sum'.

    Return:
        Tensor: value of pixel-wise binary cross entropy loss.
    """

    assert pred.shape == target.shape, \
        (f'pred shape {pred.shape} does not match target shape '
         f'{target.shape}.')

    b = pred.shape[0]

    h, w = pred.shape[2:]

    # normalize pred using sigmoid
    pred = torch.sigmoid(pred)

    if balance:
        tp_weight = 0.5 * h * w / (
            torch.sum(target, dim=[1, 2, 3]) + torch.sum(pred, dim=[1, 2, 3]) +
            1e-8)

        fn_weight = 0.5 * h * w / (
            torch.sum(1 - target, dim=[1, 2, 3]) +
            torch.sum(1 - pred, dim=[1, 2, 3]) + 1e-8)

    else:
        tp_weight, fn_weight = torch.ones(b).to(target), torch.ones(b).to(
            target)

    tp = -target * torch.log(pred + 1e-8) * tp_weight.view(b, 1, 1, 1)

    fn = -(1 - target) * torch.log(1 - pred + 1e-8) * fn_weight.view(
        b, 1, 1, 1)

    bce_map = tp + fn

    if reduction == 'none':
        return torch.squeeze(bce_map)  # shape (B, H, W).

    elif reduction == 'mean':
        return torch.mean(bce_map)

    elif reduction == 'sum':
        return torch.sum(bce_map) / b


def multi_levels_binary_cross_entropy(
    preds_dict: Dict[str, Union[Sequence[torch.Tensor], torch.Tensor]],
    target: torch.Tensor,
    weights: Dict[str, float] = dict(
        level6=0.32, level5=0.08, level4=0.02, level3=0.01, level2=0.005),
    balance: bool = False,
    reduction: str = 'sum',
) -> torch.Tensor:
    """Multi-level binary cross entropy function.

    Args:
        preds_dict (dict): multi-level output of predicted occlusion, and the
            contain in dict is a Tensor or list of Tensor with shape
            (B, 1, H_l, W_l), where l indicates the level.
        target (Tensor): ground truth of occlusion with shape (B, 1, H, W).
        weights (dict): manual rescaling weights given to the loss of occlusion
            at each level, and the keys in weights must correspond to predicted
            dict. Defaults to dict(
            level6=0.32, level5=0.08, level4=0.02, level3=0.01, level2=0.005).
        balance (bool): whether balance true positive and false negative by
            predicted and ground truth labels. Defaults to False.
        reduction (str): the reduction to apply to the output:'none', 'mean',
            'sum'. 'none': no reduction will be applied and will return a
            full-size epe map, 'mean': the mean of the epe map is taken, 'sum':
            the epe map will be summed but averaged by batch_size.
            Default: 'sum'.

    Returns:
        Tensor: value of pixel-wise binary cross entropy loss.
    """

    assert preds_dict.keys() == weights.keys(), \
        'Error: Keys of prediction do not match keys of weights!'

    loss = 0

    for k in weights.keys():
        # predict more than one flow map at one level
        cur_pred = preds_dict[k] if \
            isinstance(preds_dict[k], (tuple, list)) else [preds_dict[k]]

        num_preds = len(cur_pred)

        h, w = cur_pred[0].shape[2:]

        cur_weight = weights.get(k)

        cur_target = F.adaptive_avg_pool2d(target, [h, w])
        for pred in cur_pred:
            loss += binary_cross_entropy(pred, cur_target, balance,
                                         reduction) * cur_weight

    return loss / num_preds


@MODELS.register_module()
class MultiLevelBCE(nn.Module):
    """Multi-level binary cross entropy.

    Args:
        weights (dict): manual rescaling weights given to the loss of occlusion
            at each level, and the keys in weights must correspond to predicted
            dict. Defaults to dict(
            level6=0.32, level5=0.08, level4=0.02, level3=0.01, level2=0.005).
        balance (bool): whether balance true positive and false negative by
            predicted and ground truth labels. Defaults to False.
        reduction (str): the reduction to apply to the output:'none', 'mean',
            'sum'. 'none': no reduction will be applied and will return a
            full-size epe map, 'mean': the mean of the epe map is taken, 'sum':
            the epe map will be summed but averaged by batch_size.
            Default: 'sum'.
    """

    def __init__(self,
                 weights: Dict[str, float] = dict(
                     level6=0.32,
                     level5=0.08,
                     level4=0.02,
                     level3=0.01,
                     level2=0.005),
                 balance: bool = False,
                 reduction: str = 'sum') -> None:

        super().__init__()

        assert isinstance(balance, bool)
        self.balance = balance

        assert isinstance(weights, dict)
        self.weights = weights

        assert reduction in ('mean', 'sum')
        self.reduction = reduction

    def forward(self, preds_dict: dict, target: torch.Tensor) -> torch.Tensor:
        """Forwar function for MultiLevelBCE.

        Args:
            preds_dict (dict): multi-level output of predicted occlusion, and
                the contain in dict is a Tensor or list of Tensor with shape
                (B, 1, H_l, W_l), where l indicates the level.
            target (Tensor): ground truth of occlusion with shape (B, 1, H, W).

        Returns:
            Tensor: value of pixel-wise binary cross entropy loss.
        """

        return multi_levels_binary_cross_entropy(
            preds_dict,
            target,
            weights=self.weights,
            balance=self.balance,
            reduction=self.reduction)

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += (f'(balance={self.balance}, '
                     f'weights={self.weights}, '
                     f'reduction=\'{self.reduction}\')')
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Dict, Optional, Sequence, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptMultiConfig, OptSampleList, SampleList, TensorDict
from ..builder import build_components, build_loss
from ..utils import CorrBlock, unpack_flow_data_samples
from .base_decoder import BaseDecoder


class Upsample(nn.Module):
    """Upsampling module.

    Args:
        scale_factor (int): Scale factor of upsampling.
        channels (int): Number of channels of conv_transpose2d.
    """

    def __init__(self, scale_factor: int, channels: int) -> None:
        super().__init__()
        self.kernel_size = 2 * scale_factor - scale_factor % 2
        self.stride = scale_factor
        self.pad = math.ceil((scale_factor - 1) / 2.)
        self.channels = channels
        self.register_buffer('weight', self.bilinear_upsampling_filter())

    # caffe::BilinearFilter
    def bilinear_upsampling_filter(self) -> Tensor:
        """Generate the weights for caffe::BilinearFilter.

        Returns:
            Tensor: The weights for caffe::BilinearFilter
        """
        f = math.ceil(self.kernel_size / 2.)
        c = (2 * f - 1 - f % 2) / 2. / f
        weight = torch.zeros(self.kernel_size**2)
        for i in range(self.kernel_size**2):
            x = i % self.kernel_size
            y = (i / self.kernel_size) % self.kernel_size
            weight[i] = (1 - abs(x / f - c)) * (1 - abs(y / f - c))
        return weight.view(1, 1, self.kernel_size,
                           self.kernel_size).repeat(self.channels, 1, 1, 1)

    def forward(self, data: Tensor) -> Tensor:
        """Forward function for upsample.

        Args:
            data (Tensor): The input data.

        Returns:
            Tensor: The upsampled data.
        """
        return F.conv_transpose2d(
            data,
            self.weight,
            stride=self.stride,
            padding=self.pad,
            groups=self.channels)


class BasicBlock(BaseModule):
    """Basic convolution block for LiteFlowNet decoder modules.

    Args:
        in_channels (int): Input channels of the first convolution layer.
        feat_channels (Sequence[int]): Output channels of convolution layers in
            the block.
        conv_cfg (dict): Config of convolution layers.
        norm_cfg (dict): Config of normalization layers.
        act_cfg (dict): Config of activation layers.
        init_cfg (dict): Config of weights initialization.
    """

    def __init__(
        self,
        in_channels: int,
        feat_channels: Sequence[int],
        conv_cfg: dict,
        norm_cfg: dict,
        act_cfg: dict,
        init_cfg: dict,
    ) -> None:

        super().__init__(init_cfg)

        self.in_channels = in_channels
        self.feat_channels = feat_channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        layers = []
        for ch in self.feat_channels:
            layers.append(
                ConvModule(
                    in_channels=in_channels,
                    out_channels=ch,
                    kernel_size=3,
                    stride=1,
                    dilation=1,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg))
            in_channels = ch
        self.feat_out_channels = feat_channels[-1]
        self.layers = nn.Sequential(*layers)


class MatchingBlock(BasicBlock):
    """Matching Block.

    Args:
        corr_cfg (dict): Config for building correlation operator.
        warp_cfg (dict): Config for building warp operator.
        last_kernel_size (int): Kernel size of the last convolution layer.
        scaled_corr (bool): Whether to use scaled correlation by the number of
            elements involved to calculate correlation or not.
    """

    def __init__(
        self,
        *args,
        corr_cfg: dict,
        warp_cfg: dict,
        last_kernel_size: int,
        scaled_corr: bool,
    ) -> None:

        super().__init__(*args)
        self.corr_cfg = corr_cfg
        self.warp_op = build_components(warp_cfg)
        self.pred_flow = nn.Conv2d(
            self.feat_out_channels,
            2,
            kernel_size=last_kernel_size,
            stride=1,
            padding=last_kernel_size // 2)

        self.corr = CorrBlock(self.corr_cfg, self.act_cfg, scaled_corr)
        if corr_cfg.get('stride', 1) > 1:
            self.corr_up = Upsample(
                scale_factor=2,
                channels=(corr_cfg.get('max_displacement') * 2 + 1)**2)

        else:
            self.corr_up = nn.Sequential()

    def forward(self,
                feat1: Tensor,
                feat2: Tensor,
                upflow: Optional[Tensor] = None,
                multiplier: float = 1.) -> Tensor:
        """Forward function for MatchingBlock.

        Args:
            feat1 (Tensor): The feature from the first image.
            feat2 (Tensor): The feature from the second image.
            upflow (Optional[Tensor], optional): The upsampled optical flow
                predicted from the last level, but for level6, there is not
                optical flow from last level. Defaults to None.
            multiplier (float, optional): The constant multiplier to scale
                upsampled flow from the last level. Defaults to 1.

        Returns:
            Tensor: The optical flow predicted from MatchingBlock.
        """

        if upflow is None:
            warp_feat = feat2
            upflow = torch.zeros_like(feat1)
        else:
            warp_feat = self.warp_op(feat2, upflow * multiplier)

        corr_feat = self.corr(feat1, warp_feat)
        corr_feat = self.corr_up(corr_feat)
        feat = self.layers(corr_feat)
        res_flow = self.pred_flow(feat)

        return upflow[:, :2, ...] + res_flow


class SubpixelBlock(BasicBlock):
    """Subpixel Block.

    Args:
        warp_cfg (dict): Config dict for building warp operator.
        last_kernel_size (int): Kernel size of the last convolution layer.
    """

    def __init__(self, *args, warp_cfg: dict, last_kernel_size: int) -> None:

        super().__init__(*args)

        self.warp_op = build_components(warp_cfg)
        self.pred_flow = nn.Conv2d(
            self.feat_out_channels,
            2,
            kernel_size=last_kernel_size,
            stride=1,
            padding=last_kernel_size // 2)

    def forward(self, feat1: Tensor, feat2: Tensor, flow: Tensor,
                multiplier: float) -> Tensor:
        """Forward function for SubpixelBlock.

        Args:
            feat1 (Tensor): The feature from the first image.
            feat2 (Tensor): The feature from the second image.
            flow (Tensor): The optical flow predicted from MatchingBlock.
            multiplier (float): The constant multiplier to scale optical flow.
                Defaults to 1.

        Returns:
            Tensor:  The optical flow predicted from SubpixelBlock.
        """

        warp_feat = self.warp_op(feat2, flow * multiplier)
        feat = torch.cat((feat1, warp_feat, flow), dim=1)
        feat = self.layers(feat)
        res_flow = self.pred_flow(feat)

        return flow + res_flow


class RegularizationBlock(BasicBlock):
    """Regularization Block.

    Args:
        last_kernel_size (int, list, tuple): (List or tuple of) kernel size of
            convolution layers.
        out_channels (int): Number of output channels of convolution layers.
        warp_cfg (dict): Config dict for building warp operator.
    """

    def __init__(self, *args, last_kernel_size: Union[int, Sequence[int]],
                 out_channels: int, warp_cfg: dict) -> None:
        super().__init__(*args)

        self.warp_op = build_components(warp_cfg)
        in_channels = self.feat_out_channels

        if isinstance(last_kernel_size, (list, tuple)):
            dist_layers = []
            for ks in last_kernel_size:
                padding = [ks_ // 2 for ks_ in ks] \
                               if isinstance(ks, (tuple, list)) else ks // 2
                dist_layers.append(
                    nn.Conv2d(
                        in_channels=in_channels,
                        out_channels=out_channels,
                        kernel_size=ks,
                        padding=padding,
                        stride=1))
                in_channels = out_channels
            self.dist_layer = nn.Sequential(*dist_layers)
        else:
            self.dist_layer = nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=last_kernel_size,
                padding=last_kernel_size // 2)
        self.patch_size = int(float(out_channels)**(0.5))

    def forward(self, img1: Tensor, img2: Tensor, feat: Tensor, flow: Tensor,
                multiplier: float) -> Tensor:
        """Forward function for RegularizationBlock.

        Args:
            img1 (Tensor): The input first image.
            img2 (Tensor): The input first image.
            feat (Tensor): The input feature for RegularizationBlock.
            flow (Tensor): The optical flow predicted from RegularizationBlock.
            multiplier (float): The constant multiplier to scale optical flow.
                Defaults to 1.

        Returns:
            Tensor: The optical flow regularized by RegularizationBlock.
        """

        B = img1.shape[0]
        warp_img2 = self.warp_op(img2, flow * multiplier)
        diff_img = torch.norm((img1 - warp_img2), dim=1, p=2, keepdim=True)
        nomean_flow = flow - flow.view(B, 2, -1).mean(2, True).view(B, 2, 1, 1)
        feat = torch.cat((diff_img, nomean_flow, feat), dim=1)
        feat = self.layers(feat)

        # feature-driven distance metric
        dist_feat = self.dist_layer(feat)
        dist_feat = -dist_feat**2
        dist_feat = F.softmax(dist_feat, dim=1)

        flow_x_unfold = F.unfold(
            input=flow[:, 0:1, ...],
            kernel_size=self.patch_size,
            padding=self.patch_size // 2).view_as((dist_feat))
        flow_y_unfold = F.unfold(
            input=flow[:, 1:2, ...],
            kernel_size=self.patch_size,
            padding=self.patch_size // 2).view_as(dist_feat)

        flow_x = torch.sum(dist_feat * flow_x_unfold, dim=1, keepdim=True)
        flow_y = torch.sum(dist_feat * flow_y_unfold, dim=1, keepdim=True)

        return torch.cat((flow_x, flow_y), dim=1)


@MODELS.register_module()
class NetE(BaseDecoder):
    """NetE of LiteFlowNet.

    A sub-network structure of LiteFlowNet for cascaded flow inference and
    flow regularization.

    Args:
        in_channels (dict): Dict of input channels of different levels of
            convolution layers in feature layer.
            Default: dict(level2=32, level3=64, level4=96, level5=128,
            level6=192).
        corr_channels (dict): Dict of input channels of different levels of
            MatchingBlock (NetM).
            Default: dict(level2=144, level3=144, level4=49, level5=49,
            level6=49).
        sin_channels (dict): Dict of input channels of different levels of
            SubpixelBlock (NetS).
            Default: dict(level2=130, level3=130, level4=194, level5=258,
            level6=386).
        rin_channels (dict): Dict of input channels of different levels of
            RegularizationBlock (NetR).
            Default: dict(level2=131, level3=131, level4=131, level5=131,
            level6=195).
        feat_channels (int): Number of output channels of convolution layers in
            feature layer. Default: 64.
        mfeat_channels (tuple(int)): Tuple of output channels of convolution
            layers in MatchingBlock (NetM). Default: (128, 64, 32).
        sfeat_channels (tuple(int)): Tuple of output channels of convolution
            layers in SubpixelBlock (NetS). Default: (128, 64, 32).
        rfeat_channels (tuple(int)): Tuple of output channels of convolution
            layers in RegularizationBlock (NetR).
            Default: (128, 128, 64, 64, 32, 32).
        patch_size (dict): Dict of last kernel size of different levels
            in MatchingBlock and SubpixelBlock.
            Default: dict(level2=7, level3=5, level4=5, level5=3, level6=3).
        corr_cfg (dict): Config dict of correlation of different levels in
            MatchingBlock.
            Default: dict(
                level2=dict(type='Correlation', max_displacement=6),
                level3=dict(type='Correlation', max_displacement=6),
                level4=dict(type='Correlation', max_displacement=3),
                level5=dict(type='Correlation', max_displacement=3),
                level6=dict(type='Correlation', max_displacement=3)).
        warp_cfg (dict): Config dict for building warp operator.
            Default: dict(type='Warp').
        conv_cfg (dict, optional): Config dict of convolution layers in
            MatchingBlock, SubpixelBlock and RegularizationBlock.
            Default: None.
        norm_cfg (dict, optional): Config dict of normalization layers in
            MatchingBlock, SubpixelBlock and RegularizationBlock.
            Default: None.
        act_cfg (dict): Config dict of activation layers in MatchingBlock
            SubpixelBlock and RegularizationBlock.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        scaled_corr (bool):  Whether to use scaled correlation by the number
            of elements involved to calculate correlation or not.
            Default: True.
        regularized_flow (bool): Whether to use RegularizationBlock to
            regularize flow or not. Default: True.
        flow_loss: Config of loss function of optical flow. Default: None.
        extra_training_loss (bool): Whether to calculate the extra train loss
            for upsampled flow to the same resolution as images pair.
            Default to False.
        init_cfg (dict, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: Dict[str, int] = dict(
                     level2=32, level3=64, level4=96, level5=128, level6=192),
                 corr_channels: Dict[str, int] = dict(
                     level2=144, level3=144, level4=49, level5=49, level6=49),
                 sin_channels: Dict[str, int] = dict(
                     level2=130,
                     level3=130,
                     level4=194,
                     level5=258,
                     level6=386),
                 rin_channels: Dict[str, int] = dict(
                     level2=131,
                     level3=131,
                     level4=131,
                     level5=131,
                     level6=195),
                 feat_channels: int = 64,
                 mfeat_channels: Sequence[int] = (128, 64, 32),
                 sfeat_channels: Sequence[int] = (128, 64, 32),
                 rfeat_channels: Sequence[int] = (128, 128, 64, 64, 32, 32),
                 patch_size: Dict[str, int] = dict(
                     level2=7, level3=5, level4=5, level5=3, level6=3),
                 corr_cfg: Dict[str, dict] = dict(
                     level2=dict(type='Correlation', max_displacement=6),
                     level3=dict(type='Correlation', max_displacement=6),
                     level4=dict(type='Correlation', max_displacement=3),
                     level5=dict(type='Correlation', max_displacement=3),
                     level6=dict(type='Correlation', max_displacement=3)),
                 warp_cfg: dict = dict(type='Warp'),
                 flow_div: float = 20.,
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 scaled_corr: bool = True,
                 regularized_flow: bool = True,
                 flow_loss: Optional[dict] = None,
                 extra_training_loss: bool = False,
                 init_cfg: OptMultiConfig = None) -> None:

        super().__init__(init_cfg)

        self.in_channels = in_channels
        self.corr_channels = corr_channels
        self.sin_channels = sin_channels
        self.rin_channels = rin_channels

        self.feat_channels = feat_channels
        self.flow_levels = list(in_channels.keys())
        self.flow_levels.sort()

        self.flow_div = flow_div
        self.multiplier = dict()
        for level in self.flow_levels:
            self.multiplier[level] = self.flow_div * 2**(-int(level[-1]) + 1)

        self.start_level = self.flow_levels[-1]
        self.end_level = self.flow_levels[0]

        self.mfeat_channels = mfeat_channels
        self.sfeat_channels = sfeat_channels
        self.rfeat_channels = rfeat_channels
        self.patch_size = patch_size

        self.corr_cfg = corr_cfg
        self.warp_cfg = warp_cfg
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.scaled_corr = scaled_corr
        self.regularized_flow = regularized_flow

        if flow_loss is not None:
            self.flow_loss = build_loss(flow_loss)

        self.extra_training_loss = extra_training_loss

        self._make_layers()

    def _make_layers(self) -> None:
        """Add layers for LiteFlowNet decoder."""
        layers = []
        for level in self.flow_levels:
            in_ch = self.in_channels[level]
            corr_ch = self.corr_channels[level]
            sin_ch = self.sin_channels[level]
            rin_ch = self.rin_channels[level]
            corr_cfg = self.corr_cfg[level]
            patch_size = self.patch_size[level]

            layers.append([
                level,
                self._make_layer(level, in_ch, corr_ch, sin_ch, rin_ch,
                                 corr_cfg, patch_size)
            ])
        self.decoders = nn.ModuleDict(layers)

    def _make_layer(
        self,
        level: str,
        in_ch: int,
        corr_ch: int,
        sin_ch: int,
        rin_ch: int,
        corr_cfg: dict,
        patch_size: int,
    ) -> nn.ModuleDict:
        """Add layers at each level of LiteFlowNet decoder.

        Args:
            level (str): The level of this submodule.
            in_ch (int): The channels of input feature.
            corr_ch (int): The channels of correlation feature.
            sin_ch (int): The input channels for NetS.
            rin_ch (int): The input channels for NetR.
            corr_cfg (dict): The Config for building correlation operator.
            patch_size (int): The patch size for feature-driven local
                convolution.

        Returns:
            nn.ModuleDict: The submodule for some stage of LiteFlowNet decoder.
        """

        if in_ch < self.feat_channels:
            feat_layer = ConvModule(
                in_channels=in_ch,
                out_channels=self.feat_channels,
                kernel_size=1,
                stride=1,
                act_cfg=self.act_cfg)
        else:
            feat_layer = nn.Sequential()

        NetM = MatchingBlock(
            corr_ch,
            self.mfeat_channels,
            self.conv_cfg,
            self.norm_cfg,
            self.act_cfg,
            self.init_cfg,
            corr_cfg=corr_cfg,
            warp_cfg=self.warp_cfg,
            last_kernel_size=patch_size,
            scaled_corr=self.scaled_corr)

        NetS = SubpixelBlock(
            sin_ch,
            self.sfeat_channels,
            self.conv_cfg,
            self.norm_cfg,
            self.act_cfg,
            self.init_cfg,
            last_kernel_size=patch_size,
            warp_cfg=self.warp_cfg)
        if patch_size > 3:
            patch_kernel = [(patch_size, 1), (1, patch_size)]
        else:
            patch_kernel = patch_size

        if level != self.flow_levels[0]:
            upflow_layer = Upsample(scale_factor=2, channels=2)
        else:
            upflow_layer = nn.Sequential()

        if level != self.flow_levels[0] or self.regularized_flow:
            if in_ch < self.rfeat_channels[0]:
                rfeat_layer = ConvModule(
                    in_channels=in_ch,
                    out_channels=self.rfeat_channels[0],
                    kernel_size=1,
                    stride=1,
                    act_cfg=self.act_cfg)
            else:
                rfeat_layer = nn.Sequential()

            NetR = RegularizationBlock(
                rin_ch,
                self.rfeat_channels,
                self.conv_cfg,
                self.norm_cfg,
                self.act_cfg,
                self.init_cfg,
                last_kernel_size=patch_kernel,
                out_channels=patch_size**2,
                warp_cfg=self.warp_cfg)

            return nn.ModuleDict({
                'feat_layer': feat_layer,
                'rfeat_layer': rfeat_layer,
                'NetM': NetM,
                'NetS': NetS,
                'NetR': NetR,
                'upflow_layer': upflow_layer
            })
        else:
            return nn.ModuleDict({
                'feat_layer': feat_layer,
                'NetM': NetM,
                'NetS': NetS,
                'upflow_layer': upflow_layer
            })

    def forward(self, img1: Tensor, img2: Tensor, feat1: TensorDict,
                feat2: TensorDict) -> TensorDict:
        """Forward function for LiteFlowNet Decoder.

        Args:
            img1 (Tensor): The first input image.
            img2 (Tensor): The second input image.
            feat1 (TensorDict): The feature pyramid from first image.
            feat2 (TensorDict): The feature pyramid from second image.

        Returns:
            TensorDict: The predicted multi-level optical flow.
        """

        flow_pred = dict()

        upflow = None
        for level in self.flow_levels[::-1]:

            _feat1 = self.decoders[level]['feat_layer'](feat1[level])
            _feat2 = self.decoders[level]['feat_layer'](feat2[level])
            h, w = _feat1.shape[2:]
            _img1, _img2 = self._scale_img(img1, h,
                                           w), self._scale_img(img2, h, w)

            flowM = self.decoders[level]['NetM'](
                _feat1,
                _feat2,
                upflow,
                self.multiplier[level],
            )
            flowS = self.decoders[level]['NetS'](
                _feat1,
                _feat2,
                flowM,
                self.multiplier[level],
            )
            if level == self.end_level and not self.regularized_flow:
                upflow = self.decoders[level]['upflow_layer'](flowS)
                flow_pred[level] = flowS
            else:

                rfeat = self.decoders[level]['rfeat_layer'](feat1[level])
                flowR = self.decoders[level]['NetR'](
                    _img1,
                    _img2,
                    rfeat,
                    flowS,
                    self.multiplier[level],
                )
                upflow = self.decoders[level]['upflow_layer'](flowR)
                flow_pred[level] = flowR

        return flow_pred

    @staticmethod
    def _scale_img(img: Tensor, h: int, w: int) -> Tensor:
        return F.interpolate(
            img, size=(h, w), mode='bilinear', align_corners=False)

    def loss(self, img1: Tensor, img2: Tensor, feat1: TensorDict,
             feat2: TensorDict, data_samples: SampleList) -> TensorDict:
        """Forward function when model training.

        Args:
            img1 (Tensor): The first input image.
            img2 (Tensor): The second input image.
            feat1 (TensorDict): The feature pyramid from first image.
            feat2 (TensorDict): The feature pyramid from second image.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            TensorDict: The losses of model.
        """

        H, W = img1.shape[2:]

        flow_pred = self.forward(
            img1=img1, img2=img2, feat1=feat1, feat2=feat2)

        if self.extra_training_loss:
            flow_pred['level0'] = self._scale_img(flow_pred[self.end_level], H,
                                                  W)
        return self.loss_by_feat(flow_pred, data_samples)

    def predict(self,
                img1: Tensor,
                img2: Tensor,
                feat1: TensorDict,
                feat2: TensorDict,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function when model testing.

        Args:
            img1 (Tensor): The first input image.
            img2 (Tensor): The second input image.
            feat1 (TensorDict): The feature pyramid from first image.
            feat2 (TensorDict): The feature pyramid from second image.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        flow_pred = self.forward(
            img1=img1, img2=img2, feat1=feat1, feat2=feat2)

        flow_results = flow_pred[self.end_level]
        return self.predict_by_feat(flow_results, data_samples)

    def loss_by_feat(self, flow_pred: TensorDict,
                     data_samples: SampleList) -> TensorDict:
        """Compute optical flow loss.

        Args:
            flow_pred (Dict[str, Tensor]): multi-level predicted optical flow.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            TensorDict: The dict of losses.
        """

        loss = dict()
        batch_gt_flow_fw, _, _, _, batch_gt_valid_fw, _ = \
            unpack_flow_data_samples(data_samples)
        loss['loss_flow'] = self.flow_loss(flow_pred, batch_gt_flow_fw,
                                           batch_gt_valid_fw)
        return loss



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence, Union

import torch
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule

from mmflow.registry import MODELS


@MODELS.register_module()
class ContextNet(BaseModule):
    """The Context network to exploit contextual information for PWC to refine
    the optical flow.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels. Default: 2.
        feat_channels (Sequence[int]): List of numbers of outputs feature
            channels. Default: (128, 128, 128, 96, 64, 32).
        dilation (Sequence[int]): List of dilation of each layer. Default:
            (1, 2, 4, 8, 16, 1).
        conv_cfg (dict , optional): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict, optional): Config dict for normalization layer.
            Default: None.
        act_cfg (dict, optional): Config dict for activation layer in
            ConvModule. Default: dict(type='LeakyReLU').
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int = 2,
                 feat_channels: Sequence[int] = (128, 128, 128, 96, 64, 32),
                 dilations: Sequence[int] = (1, 2, 4, 8, 16, 1),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: Optional[Union[list, Sequence]] = None) -> None:

        super().__init__(init_cfg=init_cfg)
        assert isinstance(in_channels, int)
        layers = []
        for _feat_channels, _dilation in zip(feat_channels, dilations):
            layers.append(
                ConvModule(
                    in_channels=in_channels,
                    out_channels=_feat_channels,
                    kernel_size=3,
                    stride=1,
                    dilation=_dilation,
                    padding=_dilation,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg))
            in_channels = _feat_channels

        layers.append(
            nn.Conv2d(
                in_channels,
                out_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=True))
        self.out_channels = out_channels
        self.layers = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward function for Context network.

        Args:
            x (Tensor): Input feature.

        Returns:
            Tensor: The predicted result.
        """
        return self.layers(x)



# training schedule for pwc-net_ft schedule
train_cfg = dict(by_epoch=False, max_iters=300000, val_interval=50000)
val_cfg = dict(type='MultiValLoop')
test_cfg = dict(type='MultiTestLoop')

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='Adam', lr=3e-5, weight_decay=0.0004, betas=(0.9, 0.999)))

# learning policy
param_scheduler = dict(
    type='MultiStageLR',
    by_epoch=False,
    gammas=[0.5, 0.5],
    milestone_params=[3e-5, 2e-5],
    milestone_iters=[0, 150000],
    steps=[
        [
            45000, 65000, 85000, 95000, 97500, 100000, 110000, 120000, 130000,
            140000
        ],
        [
            195000, 215000, 235000, 245000, 247500, 250000, 260000, 270000,
            280000, 290000
        ],
    ])

default_hooks = dict(
    timer=dict(type='IterTimerHook'),
    logger=dict(type='LoggerHook', interval=50, log_metric_by_epoch=False),
    param_scheduler=dict(type='ParamSchedulerHook'),
    checkpoint=dict(type='CheckpointHook', interval=50000, by_epoch=False),
    sampler_seed=dict(type='DistSamplerSeedHook'),
    visualization=dict(type='FlowVisualizationHook'))



# Copyright (c) OpenMMLab. All rights reserved.
from .distributed_sampler import (DistributedSampler,
                                  MixedBatchDistributedSampler)

__all__ = ['DistributedSampler', 'MixedBatchDistributedSampler']



_base_ = [
    '../_base_/datasets/sintel_irr_with_occ_384x768.py',
    '../_base_/default_runtime.py',
    '../_base_/schedules/pwcnet_ft_300k_schedule.py'
]

model = dict(
    type='IRRPWC',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    encoder=dict(
        type='PWCNetEncoder',
        in_channels=3,
        net_type='Small',
        pyramid_levels=[
            'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
        ],
        out_channels=(16, 32, 64, 96, 128, 196),
        strides=(2, 2, 2, 2, 2, 2),
        dilations=(1, 1, 1, 1, 1, 1),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
    decoder=dict(
        type='IRRPWCDecoder',
        flow_levels=[
            'level0', 'level1', 'level2', 'level3', 'level4', 'level5',
            'level6'
        ],
        corr_in_channels=dict(
            level2=32, level3=64, level4=96, level5=128, level6=196),
        corr_feat_channels=32,
        flow_decoder_in_channels=115,
        occ_decoder_in_channels=114,
        corr_cfg=dict(type='Correlation', max_displacement=4),
        scaled=True,
        warp_cfg=dict(type='Warp', align_corners=True),
        densefeat_channels=(128, 128, 96, 64, 32),
        flow_post_processor=dict(
            type='ContextNet',
            in_channels=565,
            out_channels=2,
            feat_channels=(128, 128, 128, 96, 64, 32),
            dilations=(1, 2, 4, 8, 16, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
        flow_refine=dict(
            type='FlowRefine',
            in_channels=35,
            feat_channels=(128, 128, 64, 64, 32, 32),
            patch_size=3,
            warp_cfg=dict(type='Warp', align_corners=True, use_mask=True),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        occ_post_processor=dict(
            type='ContextNet',
            in_channels=563,
            out_channels=1,
            feat_channels=(128, 128, 128, 96, 64, 32),
            dilations=(1, 2, 4, 8, 16, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
        occ_refine=dict(
            type='OccRefine',
            in_channels=65,
            feat_channels=(128, 128, 64, 64, 32, 32),
            patch_size=3,
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            warp_cfg=dict(type='Warp', align_corners=True),
        ),
        occ_upsample=dict(
            type='OccShuffleUpsample',
            in_channels=11,
            feat_channels=32,
            infeat_channels=16,
            out_channels=1,
            warp_cfg=dict(type='Warp', align_corners=True, use_mask=True),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        ),
        occ_refined_levels=['level0', 'level1'],
        flow_div=20.,
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        flow_loss=dict(
            type='MultiLevelEPE',
            weights=dict(
                level6=0.32,
                level5=0.08,
                level4=0.02,
                level3=0.01,
                level2=0.005,
                level1=0.00125,
                level0=0.0003125),
            p=1,
            q=0.4,
            eps=0.01,
            reduction='sum'),
        occ_loss=dict(
            type='MultiLevelBCE',
            weights=dict(
                level6=0.32,
                level5=0.08,
                level4=0.02,
                level3=0.01,
                level2=0.005,
                level1=0.00125,
                level0=0.0003125),
            reduction='sum'),
    ),
    init_cfg=dict(
        type='Kaiming',
        a=0.1,
        nonlinearity='leaky_relu',
        layer=['Conv2d', 'ConvTranspose2d'],
        mode='fan_in',
        bias=0),
    train_cfg=dict(),
    test_cfg=dict())

custom_hooks = [dict(type='EMAHook')]

# Train on FlyingChairsOcc and FlyingThings3D_subset, and finetune on Sintel
load_from = 'https://download.openmmlab.com/mmflow/irr/irrpwc_8x1_sfine_half_flyingthings3d_subset_384x768.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod
from typing import Union

from mmengine.logging import print_log
from mmengine.model import BaseModel
from torch import Tensor, device

from mmflow.utils import (OptConfigType, OptMultiConfig, OptSampleList,
                          SampleList, TensorDict)


class FlowEstimator(BaseModel, metaclass=ABCMeta):
    """Base class for flow estimator.

    Args:
        data_preprocessor (dict or ConfigDict, optional) :The pre-process
            config for processing the input data. it usually includes
            ``bgr_to_rgb``, ``rgb_to_bgr``, ``mean`` and ``std`` for input
            images normalization and ``sigma_range`` and ``clamp_range`` for
            adding Gaussian noise on images. Defaults to None.
        freeze_net (bool): Whether freeze the weights of model. If set True,
            the model will not update the weights.
        init_cfg (dict, list, optional): Config dict of weights initialization.
            Defaults to None.
    """

    def __init__(self,
                 data_preprocessor: OptConfigType = None,
                 freeze_net: bool = False,
                 init_cfg: OptMultiConfig = None,
                 train_cfg: OptConfigType = None,
                 test_cfg: OptConfigType = None) -> None:
        super().__init__(
            data_preprocessor=data_preprocessor, init_cfg=init_cfg)

        self.status = dict(iter=0, max_iters=0)
        self.freeze_net = freeze_net
        # if set freeze_net True, the weights in this model
        # will be not updated and predict the flow maps.
        if self.freeze_net:
            print_log(
                f'Freeze the parameters in {self.__class__.__name__}',
                logger='current')
            self.eval()
            for p in self.parameters():
                p.requires_grad = False

        self.train_cfg = train_cfg
        self.test_cfg = test_cfg

    @property
    def device(self) -> device:
        """Get the current device."""
        return self.pixel_mean.device

    def forward(self,
                inputs: Tensor,
                data_samples: OptSampleList = None,
                mode: str = 'tensor') -> Union[dict, SampleList]:
        """The unified entry for a forward process in both training and test.

        The method should accept three modes: "tensor", "predict" and "loss":

        - "tensor": Forward the whole network and return tensor or tuple of
        tensor without any post-processing, same as a common nn.Module.
        - "predict": Forward and return the predictions, which are fully
        processed to a list of :obj:`FlowDataSample`.
        - "loss": Forward and return a dict of losses according to the given
        inputs and data samples.

        Note that this method doesn't handle neither back propagation nor
        optimizer updating, which are done in the :meth:`train_step`.

        Args:
            inputs (Tensor): The input tensor with shape (N, C, ...)
                in general.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.
            mode (str): Return what kind of value. Defaults to 'tensor'.

        Returns:
            The return type depends on ``mode``.

            - If ``mode="tensor"``, return a tensor or a tuple of tensor.
            - If ``mode="predict"``, return a list of :obj:`FlowDataSample`.
            - If ``mode="loss"``, return a dict of tensor.
        """

        if mode == 'loss':
            return self.loss(inputs, data_samples)
        elif mode == 'predict':
            return self.predict(inputs, data_samples)
        elif mode == 'tensor':
            return self._forward(inputs, data_samples)
        else:
            raise RuntimeError(f'Invalid mode "{mode}". '
                               'Only supports loss, predict and tensor mode')

    @abstractmethod
    def loss(self, inputs: Tensor,
             data_samples: SampleList) -> Union[dict, tuple]:
        """Calculate losses from a batch of inputs and data samples."""
        pass

    @abstractmethod
    def predict(self,
                inputs: Tensor,
                data_samples: OptSampleList = None) -> SampleList:
        """Predict results from a batch of inputs and data samples with post-
        processing."""
        pass

    @abstractmethod
    def _forward(self,
                 inputs: Tensor,
                 data_samples: OptSampleList = None) -> TensorDict:
        """Network forward process.

        Usually includes encoder,and decoder forward without any post-
        processing.
        """
        pass

    @abstractmethod
    def extract_feat(self, inputs: Tensor) -> TensorDict:
        """Extract features from images."""
        pass



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Sequence, Tuple

import torch
import torch.nn as nn
from mmcv.cnn import build_activation_layer, build_norm_layer
from mmengine.model import BaseModule
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptSampleList, SampleList, TensorDict
from ..builder import build_loss
from ..utils import unpack_flow_data_samples
from .base_decoder import BaseDecoder


class DeconvModule(BaseModule):
    """Basic deconvolution module for FlowNet.

    This module consists of several plain convolutional layers.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (int, tuple): Size of deconvolution kernel. Default: 4.
        stride (int, tuple): Whether use stride deconvolution to upsample.
            Default: 2.
        padding (int, tuple): Padding size of deconvolution. Default: 1.
        dilation (int, tuple): Dilation of deconvolution. Default: 1.
        bias (bool): Whether add bias in deconvolution layers. Default: False.
        norm_cfg (dict, optional): Config dict for normalization layer.
            Default: None.
        act_cfg (dict, optional): Config dict for activation layer.
            Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size: Tuple[int, ...] = 4,
                 stride: Tuple[int, ...] = 2,
                 padding: Tuple[int, ...] = 1,
                 dilation: Tuple[int, ...] = 1,
                 bias: bool = False,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: Optional[dict] = None) -> None:
        super().__init__()

        deconvs = []

        deconvs.append(
            nn.ConvTranspose2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                dilation=dilation,
                padding=padding,
                bias=bias))

        if norm_cfg is not None:
            deconvs.append(build_norm_layer(norm_cfg, out_channels))

        if act_cfg is not None:
            deconvs.append(build_activation_layer(act_cfg))

        self.deconvs = nn.Sequential(*deconvs)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function for deconvolution module.

        Args:
            x (Tensor): Input feature.

        Returns:
            Tensor: Output feature of deconvolution module.
        """
        out = self.deconvs(x)
        return out


class BasicBlock(BaseModule):
    """Basic block of FlowNetDecoder.

    The block includes one convolution for prediction and one
    deconvolution for next block.

    Args:
        in_channels (int): Number of input channels.
        pred_channels (int): Number of prediction channels. If 1, predict
            occlusion map. If 2, predict flow map.
        out_channels (int): Number of output channels of DeconvModule.
        inter_channels (int, Optional): Number of output channels of inter
            convolution, which is the same as the number of input channels for
            flow prediction convolution. If None, there is no inter convolution
            between input and flow prediction convolution and the input channel
            of flow prediction convolution is the same as in_channels.
            Default: None.
        deconv_bias (bool): Whether add bias term in deconvolution module.
            Default: False.
        pred_bias (bool): Whether add bias term in prediction module.
            Default: False.
        upsample_bias (bool): Whether add bias term in upsample module.
            Default: False.
        norm_cfg (dict, optional): Config dict for each normalization layer.
            Default: None.
        act_cfg (dict, optional): Config dict for each activation layer in
            ConvModule. Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 pred_channels: int,
                 out_channels: Optional[int] = None,
                 inter_channels: int = None,
                 deconv_bias: bool = False,
                 pred_bias: bool = False,
                 upsample_bias: bool = False,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: Optional[dict] = None) -> None:
        super().__init__()
        if inter_channels is None:
            self.pred_out = nn.Conv2d(
                in_channels=in_channels,
                out_channels=pred_channels,
                kernel_size=3,
                stride=1,
                padding=1,
                bias=pred_bias)
        else:
            pred_out = [
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=inter_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=pred_bias),
                nn.Conv2d(
                    in_channels=inter_channels,
                    out_channels=pred_channels,
                    kernel_size=3,
                    stride=1,
                    padding=1,
                    bias=pred_bias)
            ]
            self.pred_out = nn.Sequential(*pred_out)
        self.up_sample = out_channels is not None
        if self.up_sample:
            self.deconv_out = DeconvModule(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=4,
                stride=2,
                padding=1,
                bias=deconv_bias,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg)

            self.upsample_pred = DeconvModule(
                in_channels=pred_channels,
                out_channels=pred_channels,
                kernel_size=4,
                stride=2,
                padding=1,
                bias=upsample_bias)

    def forward(self, x: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
        """Forward function for basic block of FlowNetDecoder.

        Args:
            x (Tensor): Input feature.

        Returns:
            tuple: Predicted optical flow, upsampled optical flow and
                upsampled input feature.
        """
        upflow = None
        upfeat = None
        flow = self.pred_out(x)
        if self.up_sample:
            upfeat = self.deconv_out(x)
            upflow = self.upsample_pred(flow)

        return flow, upflow, upfeat


@MODELS.register_module()
class FlowNetSDecoder(BaseDecoder):
    """The decoder of FlowNetS.

    This module works for predicting flow and computing loss.

    Args:
        in_channels (dict): Dict of number of input channels for each level.
        out_channels (dict): Dict of number of output channels of deconvolution
            module for each level.
        pred_channels (int): Number of prediction channels. If 1, predict
            occlusion map. If 2, predict flow map. Default: 2.
        flow_div (float): The divisor works for scaling the ground truth.
            Default: 20.
        inter_channels (Sequence[int], optional): List of numbers of output
            channels of inter convolution in each BasicBlock. Check the
            description of this argument in BasicBlock for more details.
            Default: None.
        deconv_bias:  Whether add bias term in deconvolution module.
            Default: False.
        pred_bias (bool): Whether add bias term in prediction module.
            Default: False.
        upsample_bias (bool): Whether add bias term in upsample module.
            Default: False.
        norm_cfg (dict, optional): Config dict for each normalization layer.
            Default: None.
        act_cfg (dict, optional): Config dict for each activation layer in
            ConvModule. Default: dict(type='LeakyReLU', negative_slope=0.1).
        flow_loss: Config of loss function of optical flow. Default: None.
        init_cfg (dict, list, optional): Config dict of weights initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: Dict[str, int],
                 out_channels: Dict[str, int],
                 pred_channels: int = 2,
                 flow_div: float = 20.,
                 inter_channels: Optional[Sequence] = None,
                 deconv_bias: bool = False,
                 pred_bias: bool = False,
                 upsample_bias: bool = False,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 flow_loss: Optional[dict] = None,
                 init_cfg=None) -> None:

        super().__init__(init_cfg)

        self.decoders = nn.ModuleDict()
        self.flow_levels = list(in_channels.keys())
        self.flow_levels.sort()
        self.start_level = self.flow_levels[-1]
        self.end_level = self.flow_levels[0]
        self.flow_div = flow_div

        if flow_loss is not None:
            self.flow_loss = build_loss(flow_loss)

        layers = []
        for level in self.flow_levels:
            inter_ch = inter_channels.get(level) if isinstance(
                inter_channels, dict) else inter_channels
            layers.append([
                level,
                BasicBlock(
                    in_channels=in_channels[level],
                    pred_channels=pred_channels,
                    inter_channels=inter_ch,
                    out_channels=out_channels.get(level, None),
                    deconv_bias=deconv_bias,
                    pred_bias=pred_bias,
                    upsample_bias=upsample_bias,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg)
            ])
        self.decoders = nn.ModuleDict(layers)

    def forward(self, feat: TensorDict) -> TensorDict:
        """Forward function for decoder of FlowNetS.

        Args:
            feat (Dict[str, Tensor]): The feature pyramid extracted from the
                concatenated inputs.

        Returns:
            Dict[str, Tensor]: Multi-level predicted optical flow.
        """

        flow_pred = dict()
        upfeat = None
        upflow = None

        for level in self.flow_levels[::-1]:

            if level == self.start_level:
                feat_ = feat[level]
            else:
                feat_ = torch.cat((feat[level], upfeat, upflow), dim=1)

            flow, upflow, upfeat = self.decoders[level](feat_)
            flow_pred[level] = flow

        return flow_pred

    def loss(self, feat: TensorDict, data_samples: SampleList) -> TensorDict:
        """Forward function for decoder of FlowNetS when model training.

        Args:
            feat (Dict[str, Tensor]): The feature pyramid extracted from the
                concatenated inputs.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The losses of output or multi-level predicted
                optical flow.
        """

        flow_pred = self.forward(feat)

        return self.loss_by_feat(flow_pred, data_samples)

    def predict(self,
                feat: TensorDict,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function for decoder of FlowNetS when model testing.

        Args:
            feat (Dict[str, Tensor]): The feature pyramid extracted from the
                concatenated inputs.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        flow_pred = self.forward(feat)
        flow_results = flow_pred[self.end_level]

        return self.predict_by_feat(flow_results, data_samples)

    def loss_by_feat(self, flow_pred: TensorDict,
                     data_samples: SampleList) -> TensorDict:
        """The loss function for FlowNet.

        Args:
            flow_pred (Dict[str, Tensor]): multi-level predicted optical flow.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.
        Returns:
            Dict[str, Tensor]: The dict of losses.
        """
        loss = dict()
        batch_gt_flow_fw, _, _, _, batch_gt_valid_fw, _ = \
            unpack_flow_data_samples(data_samples)
        loss['loss_flow'] = self.flow_loss(flow_pred, batch_gt_flow_fw,
                                           batch_gt_valid_fw)
        return loss


@MODELS.register_module()
class FlowNetCDecoder(FlowNetSDecoder):
    """The decoder of FlowNetS."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def forward(self, feat1: TensorDict, corr_feat: TensorDict) -> TensorDict:
        """Forward function for decoder of FlowNetS.

        Args:
            feat1 (Dict[str, Tensor]): Input feature pyramid which is encoded
                image1.
            corr_feat (Dict[str, Tensor]): Input feature pyramid which is
                encoded correlation of feature1 and feature2 that are the third
                -level feature of image1 and image2.

        Returns:
            Dict[str, Tensor]: Multi-level predicted optical flow.
        """

        flow_pred = dict()
        upflow = None
        upfeat = None
        for level in self.flow_levels[::-1]:
            if level == self.start_level:
                feat = corr_feat[level]
            else:
                if corr_feat.get(level) is None:
                    feat = torch.cat((feat1[level], upfeat, upflow), dim=1)
                else:
                    feat = torch.cat((corr_feat[level], upfeat, upflow), dim=1)

            flow, upflow, upfeat = self.decoders[level](feat)

            flow_pred[level] = flow

        return flow_pred

    def loss(self, feat1: TensorDict, corr_feat: TensorDict,
             data_samples: SampleList) -> TensorDict:
        """Forward function for decoder of FlowNetS when model training.

        Args:
            feat1 (Dict[str, Tensor]): Input feature pyramid which is encoded
                image1.
            corr_feat (Dict[str, Tensor]): Input feature pyramid which is
                encoded correlation of feature1 and feature2 that are the third
                -level feature of image1 and image2.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The losses of output or multi-level predicted
                optical flow.
        """

        flow_pred = self.forward(feat1, corr_feat)

        return self.loss_by_feat(flow_pred, data_samples)

    def predict(self,
                feat1: TensorDict,
                corr_feat: TensorDict,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function for decoder of FlowNetS when model testing.

        Args:
            feat1 (Dict[str, Tensor]): Input feature pyramid which is encoded
                image1.
            corr_feat (Dict[str, Tensor]): Input feature pyramid which is
                encoded correlation of feature1 and feature2 that are the third
                -level feature of image1 and image2.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        flow_pred = self.forward(feat1, corr_feat)
        flow_results = flow_pred[self.end_level]

        return self.predict_by_feat(flow_results, data_samples)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import json
import os.path as osp

from mmengine.utils import mkdir_or_exist
from utils import get_data_filename


def parse_args():
    parser = argparse.ArgumentParser(
        description='ChairsSDHom dataset preparation')
    parser.add_argument(
        '--data-root',
        type=str,
        default='data/ChairsSDHom',
        help='Directory for dataset.')

    parser.add_argument(
        '--save-dir',
        type=str,
        default='data/ChairsSDHom/',
        help='Directory to save the annotation files for ChairsSDHom dataset')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()

    # these files contain nan, so exclude them.
    exclude_files = ['08755.pfm']

    img1_suffix = '.png'
    img2_suffix = '.png'
    flow_suffix = '.pfm'

    def _get_data_list(subset_dir):
        data_root = osp.join(args.data_root, 'data', subset_dir)
        img1_dir = osp.join(data_root, 't0')
        img2_dir = osp.join(data_root, 't1')
        flow_dir = osp.join(data_root, 'flow')

        flow_filenames = get_data_filename(
            flow_dir, flow_suffix, exclude=exclude_files)
        img1_filenames, img2_filenames = _revise_dir(flow_filenames, img1_dir,
                                                     img2_dir)

        data_list = []

        for i in range(len(flow_filenames)):
            data_info = dict(
                img1_path=img1_filenames[i],
                img2_path=img2_filenames[i],
                flow_fw_path=flow_filenames[i])
            data_list.append(data_info)
        mkdir_or_exist(args.save_dir)
        with open(osp.join(args.save_dir, f'{subset_dir}.json'),
                  'w') as jsonfile:
            json.dump({
                'metainfo': {},
                'data_list': data_list
            },
                      jsonfile,
                      indent=4)

    def _revise_dir(flow_filenames, img1_dir, img2_dir):
        img1_filenames = []
        img2_filenames = []
        for flow_filename in flow_filenames:
            idx = int(osp.splitext(osp.basename(flow_filename))[0])
            img1_filename = osp.join(img1_dir, f'{idx:05d}' + img1_suffix)
            img2_filename = osp.join(img2_dir, f'{idx:05d}' + img2_suffix)
            img1_filenames.append(img1_filename)
            img2_filenames.append(img2_filename)
        return img1_filenames, img2_filenames

    _get_data_list('train')
    _get_data_list('test')


if __name__ == '__main__':
    main()



# Copyright (c) OpenMMLab. All rights reserved.
import copy
import warnings
from math import ceil
from typing import Sequence, Tuple, Union

import mmcv
import numpy as np
from mmcv.image import adjust_brightness, adjust_color, adjust_contrast
from mmcv.transforms import BaseTransform
from mmcv.transforms.utils import cache_randomness
from numpy import random

from mmflow.registry import TRANSFORMS
from ..utils import adjust_gamma, adjust_hue

img_keys = ['img1', 'img2']

flow_keys = ['gt_flow_fw', 'gt_flow_bw']

occ_keys = ['gt_occ_fw', 'gt_occ_bw']

valid_keys = ['gt_valid_fw', 'gt_valid_bw']


@TRANSFORMS.register_module()
class SpacialTransform(BaseTransform):
    """Spacial Transform API for RAFT.

    Required Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)


    Modified Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)


    Added Keys:

    - scale

    Args:
        spacial_prob (float): probability to do spacial transform.
        stretch_prob (float): probability to do stretch.
        crop_size (tuple, list): the base size for resize.
        min_scale (float): the exponent for min scale. Defaults to -0.2.
        max_scale (float): the exponent for max scale. Defaults to 0.5.
    Returns:
        dict: Resized results, 'img_shape',
    """

    def __init__(self,
                 spacial_prob: float,
                 stretch_prob: float,
                 crop_size: Sequence,
                 min_scale: float = -0.2,
                 max_scale: float = 0.5,
                 max_stretch: float = 0.2) -> None:
        super().__init__()
        assert spacial_prob >= 0. and spacial_prob <= 1. and isinstance(
            spacial_prob, float)
        assert stretch_prob >= 0. and stretch_prob <= 1. and isinstance(
            stretch_prob, float)
        assert isinstance(
            crop_size, (tuple, list)) and len(crop_size) == 2 and isinstance(
                crop_size[0], int) and isinstance(crop_size[1], int)
        assert isinstance(min_scale, float)
        assert isinstance(max_scale, float)
        assert isinstance(max_stretch, float)
        self.spacial_prob = spacial_prob
        self.stretch_prob = stretch_prob
        self.crop_size = crop_size
        self.min_scale = min_scale
        self.max_scale = max_scale
        self.max_stretch = max_stretch

    def transform(self, results: dict) -> dict:
        """Call function to do spacial transform to images and annotation,
        including optical flow, occlusion mask and valid mask.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.

        Returns:
            dict: The dict contains transformed data and transform information.
        """

        if self._do_spacial_transform():

            H, W = results['img1'].shape[:2]
            self.newH, self.newW, self.x0, self.y0 = self._random_scale(H, W)

            if not results['sparse']:
                self._dense_flow_transform(results)
            else:
                self._sparse_flow_transform(results)
        else:
            results['scale'] = (1., 1.)

        return results

    @cache_randomness
    def _do_spacial_transform(self) -> bool:
        """Whether do spacial transform.

        Returns:
            bool: If True, images and flow will do spacial transform.
        """
        return np.random.rand() < self.spacial_prob

    @cache_randomness
    def _random_scale(self, H, W):
        """Sample image scale randomly.

        Args:
            H (int): The height of input images.
            W (int): The width of input images

        Returns:
            Tuple(int): The new height, width and coordinates for crop box.
        """
        min_scale = np.maximum((self.crop_size[0] + 8) / float(H),
                               (self.crop_size[1] + 8) / float(W))
        scale = 2**np.random.uniform(self.min_scale, self.max_scale)
        scale_x = scale
        scale_y = scale
        if np.random.rand() < self.stretch_prob:
            scale_x *= 2**np.random.uniform(-self.max_stretch,
                                            self.max_stretch)
            scale_y *= 2**np.random.uniform(-self.max_stretch,
                                            self.max_stretch)
        scale_x = np.clip(scale_x, min_scale, None)
        scale_y = np.clip(scale_y, min_scale, None)
        newW = int(W * float(scale_x) + 0.5)
        newH = int(H * float(scale_y) + 0.5)
        y0 = np.random.randint(0, newH - self.crop_size[0])
        x0 = np.random.randint(0, newW - self.crop_size[1])
        return newH, newW, x0, y0

    def _dense_flow_transform(self, results):
        """Transform for dense flow map.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.
        """
        # transform images
        for k in img_keys:
            if results.get(k) is not None:
                results[k], scale_x, scale_y, = self._resize_crop(results[k])
        results['scale'] = (scale_x, scale_y)
        results['img_shape'] = results['img1'].shape

        # transform flows
        for k in flow_keys:
            if results.get(k) is not None:
                flow, scale_x, scale_y = self._resize_crop(results[k])
                flow *= [scale_x, scale_y]
                results[k] = flow

        # transform occ
        for k in occ_keys:
            if results.get(k, None) is not None:
                results[k], _, _ = self._resize_crop(results[k])

        # transform valid
        for k in valid_keys:
            if results.get(k, None) is not None:
                results[k], _, _ = self._resize_crop(results[k])

    def _resize_crop(
        self,
        img: np.ndarray,
        interpolation: str = 'bilinear'
    ) -> Tuple[np.ndarray, float, float, int, int]:
        """Spacial transform function.

        Args:
            img (ndarray): the images that will be transformed.

        Returns:
            Tuple[ndarray, float, float, int, int]: the transformed images,
                horizontal scale factor, vertical scale factor, coordinate of
                left-top point where the image maps will be crop from.
        """

        img_, scale_x, scale_y = mmcv.imresize(
            img, (self.newW, self.newH),
            return_scale=True,
            interpolation=interpolation)
        img_ = img_[self.y0:self.y0 + self.crop_size[0],
                    self.x0:self.x0 + self.crop_size[1]]

        return img_, scale_x, scale_y

    def _sparse_flow_transform(self, results):
        """Transform for sparse flow map.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.
        """
        # sparse spacial_transform for kitti/hd1k dataset
        # transform images
        for k in img_keys:
            if results.get(k) is not None:
                results[k], scale_x, scale_y, = self._resize_crop(results[k])
        results['scale'] = (scale_x, scale_y)
        results['img_shape'] = results['img1'].shape

        # transform flow_fw and valid
        flow, valid = self._sparse_flow_resize_crop(
            results['gt_flow_fw'],
            results['gt_valid_fw'],
            fx=scale_x,
            fy=scale_y)

        results['gt_flow_fw'] = flow
        results['gt_valid_fw'] = valid.astype(np.float32)

    def _sparse_flow_resize_crop(self,
                                 flow: np.ndarray,
                                 valid: np.ndarray,
                                 fx: float = 1.0,
                                 fy: float = 1.0) -> Sequence[np.ndarray]:
        """Resize sparse optical flow function.

        Args:
            flow (ndarray): optical flow data will be resized.
            valid (ndarray): valid mask for sparse optical flow.
            fx (float, optional): horizontal scale factor. Defaults to 1.0.
            fy (float, optional): vertical scale factor. Defaults to 1.0.

        Returns:
            Sequence[ndarray]: the transformed flow map and valid mask.
        """
        ht, wd = flow.shape[:2]
        coords = np.meshgrid(np.arange(wd), np.arange(ht))
        coords = np.stack(coords, axis=-1)

        coords = coords.reshape(-1, 2).astype(np.float32)
        flow = flow.reshape(-1, 2).astype(np.float32)
        valid = valid.reshape(-1).astype(np.float32)

        coords0 = coords[valid >= 1]
        flow0 = flow[valid >= 1]

        ht1 = int(round(ht * fy))
        wd1 = int(round(wd * fx))

        coords1 = coords0 * [fx, fy]
        flow1 = flow0 * [fx, fy]

        xx = np.round(coords1[:, 0]).astype(np.int32)
        yy = np.round(coords1[:, 1]).astype(np.int32)

        v = (xx > 0) & (xx < wd1) & (yy > 0) & (yy < ht1)
        xx = xx[v]
        yy = yy[v]
        flow1 = flow1[v]

        flow_img = np.zeros([ht1, wd1, 2], dtype=np.float32)
        valid_img = np.zeros([ht1, wd1], dtype=np.int32)

        flow_img[yy, xx] = flow1
        valid_img[yy, xx] = 1.
        flow_img = flow_img[self.y0:self.y0 + self.crop_size[0],
                            self.x0:self.x0 + self.crop_size[1]]
        valid_img = valid_img[self.y0:self.y0 + self.crop_size[0],
                              self.x0:self.x0 + self.crop_size[1]]
        return flow_img, valid_img

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += (f'(spacial_prob={self.spacial_prob} '
                     f'stretch_prob={self.stretch_prob} '
                     f'crop_size={self.crop_size} '
                     f'min_scale={self.min_scale} '
                     f'max_scale = {self.max_scale}')
        return repr_str


@TRANSFORMS.register_module()
class Validation(BaseTransform):
    """This Validation transform from RAFT is for return a mask for the flow is
    less than max_flow.

    Required Keys:

    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)


    Modified Keys:

    - gt_valid_fw (optional)
    - gt_valid_bw (optional)

    Args:
        max_flow (float, int): the max flow for validated flow.
    Returns:
        dict: Resized results, 'valid' and 'max_flow' keys are added into
            result dict.
    """

    def __init__(self, max_flow: Union[float, int]) -> None:
        assert isinstance(max_flow, (float, int))
        self.max_flow = max_flow

    def transform(self, results: dict) -> dict:
        """Call function to get the valid mask.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.

        Returns:
            dict: dict added 'valid' key and its value.
        """

        for k in flow_keys:
            if results.get(k, None) is not None:
                flow = results[k]
                valid = ((np.abs(flow[:, :, 0]) < self.max_flow) &
                         (np.abs(flow[:, :, 1]) < self.max_flow))
                valid_key = k.replace('flow', 'valid')
                if results.get(valid_key, None) is not None:
                    results[valid_key] = (
                        valid.astype(np.float32) * results[valid_key])
                else:
                    results[valid_key] = valid.astype(np.float32)
        results['max_flow'] = self.max_flow
        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(max_flow={self.img_scale})'
        return repr_str


@TRANSFORMS.register_module()
class Erase(BaseTransform):
    """Erase transform from RAFT is randomly erasing rectangular regions in
    img2 to simulate occlusions.

    Required Keys:

    - img2

    Modified Keys:

    - img2

    Added  Keys:

    - erase_num
    - erase_bounds

    Args:
        prob (float): the probability for erase transform.
        bounds (list, tuple): the bounds for erase regions (bound_x, bound_y).
        max_num (int): the max number of erase regions.

    Returns:
        dict: revised results, 'img2' and 'erase_num' are added into results.
    """

    def __init__(self,
                 prob: float,
                 bounds: Sequence = [50, 100],
                 max_num: int = 3) -> None:
        assert isinstance(prob, float), ('Probability for erase transform must'
                                         f' be float, but got {type(prob)}')
        assert prob >= 0 and prob <= 1, ('The range of probability is [0.,1.],'
                                         f' but got {type(prob)}')

        assert isinstance(max_num,
                          int), f'max_num must be int, but got {type(max_num)}'
        self.prob = prob
        self.bounds = bounds
        self.max_num = max_num

    @cache_randomness
    def _do_erase(self):
        """Whether do erase transform.

        Returns:
            bool: If True, do this transform.
        """
        return np.random.rand() < self.prob

    def transform(self, results: dict) -> dict:
        """Call function to do erase on images.

        Args:
            results (dict): Result dict from :obj:`mmflow.BaseDataset`.

        Returns:
            dict: the values of 'img1' and 'img2' is updated, and add
                'erase_num' and 'erase_bounds' keys and their values.
        """
        img2 = results['img2']
        H, W, _ = img2.shape
        erase_bounds = []
        num = 0
        if self._do_erase():
            mean_color = np.mean(img2.reshape(-1, 3), axis=0)
            num = np.random.randint(1, self.max_num)
            for _ in range(num):
                x0 = np.random.randint(0, W)
                y0 = np.random.randint(0, H)
                dx = np.random.randint(self.bounds[0], self.bounds[1])
                dy = np.random.randint(self.bounds[0], self.bounds[1])
                img2[y0:y0 + dy, x0:x0 + dx, :] = mean_color
                erase_bounds.append((y0, x0, y0 + dy, x0 + dx))
        results['img2'] = img2
        results['erase_num'] = num
        results['erase_bounds'] = erase_bounds

        return results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += (f'(prob={self.prob} '
                     f'bounds={self.bounds} '
                     f'max_num={self.max_num})')

        return repr_str


@TRANSFORMS.register_module()
class InputResize(BaseTransform):
    """Resize images such that dimensions are divisible by 2^n.

    Required Keys:

    - img1
    - img2

    Modified Keys:

    - img1
    - img2
    - img_shape

    Added Keys:

    - scale_factor


    Args:
        exponent(int): the exponent n of 2^n

    Returns:
        dict: Resized results, 'img_shape', 'scale_factor' keys are added
            into result dict.
    """

    def __init__(self, exponent: int) -> None:
        super().__init__()
        assert isinstance(exponent, int)
        self.exponent = exponent

    def transform(self, results: dict) -> dict:
        """Call function to resize images and flow map.

        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Resized results, 'img_shape', 'scale_factor' keys are added
                into result dict.
        """
        self._resize_img(results)

        return results

    def _resize_img(self, results: dict) -> None:
        """Resize images with ``results['scale']``."""
        img1 = results['img1']
        img2 = results['img2']
        times = int(2**self.exponent)
        H, W = img1.shape[:2]
        newH = int(ceil(H / times) * times)
        newW = int(ceil(W / times) * times)
        resized_img1 = mmcv.imresize(img1, (newW, newH), return_scale=False)
        resized_img2 = mmcv.imresize(img2, (newW, newH), return_scale=False)
        results['img1'] = resized_img1
        results['img2'] = resized_img2

        w_scale = newW / W
        h_scale = newH / H
        scale_factor = np.array([w_scale, h_scale], dtype=np.float32)
        results['scale_factor'] = scale_factor
        results['img_shape'] = resized_img1.shape

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += (f'(exponent={self.exponent})')
        return repr_str


@TRANSFORMS.register_module()
class InputPad(BaseTransform):
    """Pad images such that dimensions are divisible by 2^n used in test.

    Required Keys:

    - img1
    - img2

    Modified Keys:

    - img1
    - img2
    - img_shape

    Added Keys:

    - pad_shape
    - pad

    Args:
        exponent(int): the exponent n of 2^n
        mode(str): mode for numpy.pad(). Defaults to 'edge'.
        position(str) the position of origin image, and valid value is one of
            'center', 'left', 'right', 'top' and 'down'. Defaults to 'center'
    """

    def __init__(self,
                 exponent: int,
                 mode: str = 'edge',
                 position: str = 'center',
                 **kwargs) -> None:
        assert position in ('center', 'left', 'right', 'top', 'down')
        assert isinstance(exponent, int)
        self.exponent = exponent
        self.mode = mode
        self.position = position
        self.kwargs = kwargs

    def transform(self, results: dict) -> dict:
        """Call function to pad image.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            dict: Updated result dict.
        """
        self._pad_img(results)

        return results

    def _pad_img(self, results: dict) -> None:
        """Pad images according to ``self.exponent``."""

        img1 = results['img1']
        img2 = results['img2']
        times = int(2**self.exponent)
        H, W = img1.shape[:2]
        pad_h = (((H // times) + 1) * times - H) % times
        pad_w = (((W // times) + 1) * times - W) % times
        if self.position == 'center':
            self._pad = [[pad_h // 2, pad_h - pad_h // 2],
                         [pad_w // 2, pad_w - pad_w // 2]]
        elif self.position == 'left':
            self._pad = [[pad_h // 2, pad_h - pad_h // 2], [0, pad_w]]
        elif self.position == 'right':
            self._pad = [[pad_h // 2, pad_h - pad_h // 2], [pad_w, 0]]
        elif self.position == 'top':
            self._pad = [[0, pad_h, pad_w // 2], [pad_w - pad_w // 2]]
        elif self.position == 'down':
            self._pad = [[pad_h, 0], [pad_w // 2, pad_w - pad_w // 2]]
        if len(img1.shape) > 2:
            self._pad.append([0, 0])
        padded_img1 = np.pad(img1, self._pad, mode=self.mode, **self.kwargs)
        padded_img2 = np.pad(img2, self._pad, mode=self.mode, **self.kwargs)
        results['img1'] = padded_img1
        results['img2'] = padded_img2
        results['pad_shape'] = padded_img1.shape
        results['img_shape'] = padded_img1.shape
        results['pad'] = self._pad[:2]

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += (f'(exponent={self.exponent} '
                     f'mode={self.mode} '
                     f'position={self.position})')
        return repr_str


@TRANSFORMS.register_module()
class RandomFlip(BaseTransform):
    """Flip the image and flow map.

    Required Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)


    Modified Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)

    Added Keys:

    - flip
    - flip_direction

    Args:
        prob (float): The flipping probability.
        direction(str): The flipping direction. Options are
            'horizontal' and 'vertical'. Default: 'horizontal'.
    """

    def __init__(self, prob, direction='horizontal'):
        assert isinstance(prob, (float, int)) and prob >= 0 and prob <= 1
        assert direction in ['horizontal', 'vertical']
        self.prob = prob
        self.direction = direction

    @cache_randomness
    def do_flip(self):
        return np.random.rand() < self.prob

    def transform(self, results):
        """Call function to flip optical flow map.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            dict: Flipped results, 'flip', 'flip_direction' keys are added into
                result dict.
        """
        if self.do_flip():
            self._flip(results)
            if 'flip' in results and 'flip_direction' in results:
                results['flip'].append(True)
                results['flip_direction'].append(self.direction)
            else:
                results['flip'] = [True]
                results['flip_direction'] = [self.direction]
        else:
            if 'flip' in results and 'flip_direction' in results:
                results['flip'].append(False)
                results['flip_direction'].append(None)
            else:
                results['flip'] = [False]
                results['flip_direction'] = [None]

        return results

    def _flip(self, results: dict) -> None:
        """Flip function for images and annotations.

        Args:
            results (dict): Result dict from loading pipeline.
        """

        # flip img
        for k in img_keys:
            results[k] = mmcv.imflip(results[k], direction=self.direction)

        # flip flow
        if self.direction == 'horizontal':
            coeff = [-1, 1]
        else:
            coeff = [1, -1]
        for k in flow_keys:
            if results.get(k, None) is not None:
                results[k] = mmcv.imflip(
                    results[k], direction=self.direction) * coeff
        # flip occ
        for k in occ_keys:
            if results.get(k, None) is not None:
                results[k] = mmcv.imflip(results[k], direction=self.direction)

        # flip valid mask
        for k in valid_keys:
            if results.get(k, None) is not None:
                results[k] = mmcv.imflip(
                    results[k], direction=self.direction).copy()

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__ + f'(prob={self.prob}), '
        repr_str += f'(direction={self.direction})'
        return repr_str


@TRANSFORMS.register_module()
class Normalize(BaseTransform):
    """Normalize the image.

    Added key is "img_norm_cfg".
    Args:
        mean (sequence): Mean values of 3 channels.
        std (sequence): Std values of 3 channels.
        to_rgb (bool): Whether to convert the image from BGR to RGB,
            default is true.
    """

    def __init__(self, mean, std, to_rgb=True):
        self.mean = np.array(mean, dtype=np.float32)
        self.std = np.array(std, dtype=np.float32)
        self.to_rgb = to_rgb

    def transform(self, results):
        """Call function to normalize images.

        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Normalized results, 'img_norm_cfg' key is added into
                result dict.
        """
        if results.get('channels_order') == 'RGB':
            self.to_rgb = False
            warnings.warn('The channels order is RBG, '
                          'and image will not convert it again')
        for k in img_keys:
            results[k] = mmcv.imnormalize(results[k], self.mean, self.std,
                                          self.to_rgb)
        results['img_norm_cfg'] = dict(
            mean=self.mean, std=self.std, to_rgb=self.to_rgb)
        if self.to_rgb:
            results['channels_order'] = 'RGB'
        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(mean={self.mean}, std={self.std}, to_rgb=' \
                    f'{self.to_rgb})'
        return repr_str


@TRANSFORMS.register_module()
class BGR2RGB(BaseTransform):
    """Convert image channels from BGR to RGB order.

    Returns:
        dict: results contained converted images.
    """

    def __init__(self):
        super().__init__()

    def transform(self, results):
        for k in img_keys:
            results[k] = mmcv.bgr2rgb(results[k])
        results['channels_order'] = 'RGB'
        return results


@TRANSFORMS.register_module()
class Rerange(BaseTransform):
    """Rerange the image pixel value.

    Args:
        min_value (float or int): Minimum value of the reranged image.
            Default: 0.
        max_value (float or int): Maximum value of the reranged image.
            Default: 255.
    """

    def __init__(self, min_value=0, max_value=255):
        assert isinstance(min_value, float) or isinstance(min_value, int)
        assert isinstance(max_value, float) or isinstance(max_value, int)
        assert min_value < max_value
        self.min_value = min_value
        self.max_value = max_value

    def transform(self, results):
        """Call function to rerange images.

        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Reranged results.
        """
        for k in img_keys:
            img = results[k]
            img_min_value = np.min(img)
            img_max_value = np.max(img)
            assert img_min_value < img_max_value
            # rerange to [0, 1]
            img = (img - img_min_value) / (img_max_value - img_min_value)
            # rerange to [min_value, max_value]
            img = img * (self.max_value - self.min_value) + self.min_value
            results[k] = img

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += f'(min_value={self.min_value}, max_value={self.max_value})'
        return repr_str


@TRANSFORMS.register_module()
class RandomCrop(BaseTransform):
    """Random crop the image & flow.

    Required Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)


    Modified Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)
    - img_shape

    Added Keys:

    - crop_bbox

    Args:
        crop_size (tuple): Expected size after cropping, (h, w).
    """

    def __init__(self, crop_size):
        assert crop_size[0] > 0 and crop_size[1] > 0
        self.crop_size = crop_size

    @cache_randomness
    def get_crop_bbox(self, img_shape: tuple) -> tuple:
        """Randomly get a crop bounding box.

        Args:
            img_shape (tuple): The shape of images

        Returns:
            tuple: The crop box for images cropping.s
        """
        margin_h = max(img_shape[0] - self.crop_size[0], 0)
        margin_w = max(img_shape[1] - self.crop_size[1], 0)
        offset_h = np.random.randint(0, margin_h + 1)
        offset_w = np.random.randint(0, margin_w + 1)
        crop_y1, crop_y2 = offset_h, offset_h + self.crop_size[0]
        crop_x1, crop_x2 = offset_w, offset_w + self.crop_size[1]

        return crop_y1, crop_y2, crop_x1, crop_x2

    def crop(self, img, crop_bbox):
        """Crop from ``img``"""
        crop_y1, crop_y2, crop_x1, crop_x2 = crop_bbox
        img = img[crop_y1:crop_y2, crop_x1:crop_x2, ...]
        return img

    def transform(self, results):
        """Call function to randomly crop images, flow maps.

        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Randomly cropped results, 'img_shape' key in result dict is
                updated according to crop size.
        """
        img_shape = copy.deepcopy(results['img_shape'])
        crop_bbox = self.get_crop_bbox(img_shape)

        # crop imgs
        for k in img_keys:
            if results.get(k, None) is not None:
                results[k] = self.crop(results[k], crop_bbox=crop_bbox)

        # crop flow
        for k in flow_keys:
            if results.get(k, None) is not None:
                results[k] = self.crop(results[k], crop_bbox=crop_bbox)

        # crop occ
        for k in occ_keys:
            if results.get(k, None) is not None:
                results[k] = self.crop(results[k], crop_bbox=crop_bbox)

        # crop valid
        for k in valid_keys:
            if results.get(k, None) is not None:
                results[k] = self.crop(results[k], crop_bbox=crop_bbox)

        results['img_shape'] = results['img1'].shape
        results['crop_bbox'] = crop_bbox

        return results

    def __repr__(self):
        return self.__class__.__name__ + f'(crop_size={self.crop_size})'


@TRANSFORMS.register_module()
class ColorJitter(BaseTransform):
    """Randomly change the brightness, contrast, saturation and hue of an
    image.

    Required Keys:

    - img1
    - img2

    Modified Keys:

    - img1
    - img2

    Args:
        asymmetric_prob (float): the probability to do color jitter for two
            images asymmetrically.
        brightness (float, tuple):  How much to jitter brightness.
            brightness_factor is chosen uniformly from
            [max(0, 1 - brightness), 1 + brightness] or the given [min, max].
            Should be non negative numbers.
        contrast (float, tuple):  How much to jitter contrast.
            contrast_factor is chosen uniformly from
            [max(0, 1 - contrast), 1 + contrast] or the given [min, max].
            Should be non negative numbers.
        saturation (float, tuple):  How much to jitter saturation.
            saturation_factor is chosen uniformly from
            [max(0, 1 - saturation), 1 + saturation] or the given [min, max].
            Should be non negative numbers.
        hue (float, tuple): How much to jitter hue.
            hue_factor is chosen uniformly from [-hue, hue] or the given
            [min, max]. Should have 0<= hue <= 0.5 or
            -0.5 <= min <= max <= 0.5.
    """

    def __init__(self,
                 asymmetric_prob=0.,
                 brightness=0.,
                 contrast=0.,
                 saturation=0.,
                 hue=0.):
        assert isinstance(
            asymmetric_prob, float
        ), f'asymmetric_prob must be float, but got {type(asymmetric_prob)}'
        self.asymmetric_prob = asymmetric_prob

        self._brightness = self._check_input(brightness, 'brightness')
        self._contrast = self._check_input(contrast, 'contrast')
        self._saturation = self._check_input(saturation, 'saturation')
        self._hue = self._check_input(
            hue, 'hue', center=0., bound=(-0.5, 0.5), clip_first_on_zero=False)

    def _get_param(self):

        fn_idx = np.random.permutation(4)
        b = None if self._brightness is None else np.random.uniform(
            self._brightness[0], self._brightness[1])
        c = None if self._contrast is None else np.random.uniform(
            self._contrast[0], self._contrast[1])
        s = None if self._saturation is None else np.random.uniform(
            self._saturation[0], self._saturation[1])
        h = None if self._hue is None else np.random.uniform(
            self._hue[0], self._hue[1])

        return fn_idx, b, c, s, h

    def _check_input(self,
                     value,
                     name,
                     center=1,
                     bound=(0, float('inf')),
                     clip_first_on_zero=True):
        if isinstance(value, (float, int)):

            if value < 0:
                raise ValueError(
                    f'If {name} is a single number, it must be non negative.')
            value = [center - float(value), center + float(value)]
            if clip_first_on_zero:
                value[0] = max(value[0], 0.0)
            if not bound[0] <= value[0] <= value[1] <= bound[1]:
                raise ValueError(f'{name} values should be between {bound}')

        elif isinstance(value, (tuple, list)) and len(value) == 2:

            if not bound[0] <= value[0] <= value[1] <= bound[1]:
                raise ValueError(f'{name} values should be between {bound}')

        else:
            raise TypeError(
                f'{name} should be a single number or a list/tuple with '
                f'length 2, but got {value}.')

        # if value is 0 or (1., 1.) for brightness/contrast/saturation
        # or (0., 0.) for hue, do nothing
        if value[0] == value[1] == center:
            value = None
        return value

    def color_jitter(self, img):
        fn_idx, brightness, contrast, saturation, hue = self._get_param()

        img = img if isinstance(img, (list, tuple)) else [img]
        length = len(img)
        for i in fn_idx:
            if i == 0 and brightness:
                img = [adjust_brightness(i, brightness) for i in img]

            if i == 1 and contrast:
                img = [adjust_contrast(i, contrast) for i in img]

            if i == 2 and saturation:
                img = [adjust_color(i, saturation) for i in img]

            if i == 3 and hue:
                img = [adjust_hue(i, hue) for i in img]
        if length == 1:
            return img[0]
        return img

    @cache_randomness
    def _do_asymmetric(self) -> bool:
        """Random function for asymmetric.

        Returns:
            bool: Whether do color jitter for images asymmetrically.
        """

        return np.random.rand() < self.asymmetric_prob

    def transform(self, results):
        """Call function to perform photometric distortion on images.

        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Result dict with images distorted.
        """

        imgs = []
        for k in img_keys:
            imgs.append(results[k])

        # asymmetric
        if self._do_asymmetric():
            imgs_ = []
            for i in imgs:
                i = self.color_jitter(i)
                imgs_.append(i)
            imgs = imgs_
        else:
            # symmetric
            imgs = self.color_jitter(imgs)
        for i, k in enumerate(img_keys):
            results[k] = imgs[i]

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += (f'asymmetric_prob={self.asymmetric_prob}, '
                     f'brightness_range={self._brightness}, '
                     f'contrast_range={self._contrast}, '
                     f'saturation_range={self._saturation}, '
                     f'hue_range={self._hue}')
        return repr_str


@TRANSFORMS.register_module()
class PhotoMetricDistortion(BaseTransform):
    """Apply photometric distortion to image sequentially, every transformation
    is applied with a probability of 0.5.

    The position of random contrast is in
    second or second to last.
    1. random brightness
    2. random contrast (mode 0)
    3. convert color from BGR to HSV
    4. random saturation
    5. random hue
    6. convert color from HSV to BGR
    7. random contrast (mode 1)
    8. randomly swap channels
    Args:
        brightness_delta (int): delta of brightness.
        contrast_range (tuple): range of contrast.
        saturation_range (tuple): range of saturation.
        hue_delta (int): delta of hue.
    """

    def __init__(self,
                 brightness_delta=32,
                 contrast_range=(0.5, 1.5),
                 saturation_range=(0.5, 1.5),
                 hue_delta=18):
        self.brightness_delta = brightness_delta
        self.contrast_lower, self.contrast_upper = contrast_range
        self.saturation_lower, self.saturation_upper = saturation_range
        self.hue_delta = hue_delta

    def convert(self, img, alpha=1, beta=0):
        """Multiple with alpha and add beat with clip."""
        img = img.astype(np.float32) * alpha + beta
        img = np.clip(img, 0, 255)
        return img.astype(np.uint8)

    def brightness(self, img):
        """Brightness distortion."""
        if random.randint(2):
            beta = random.uniform(-self.brightness_delta,
                                  self.brightness_delta)
            img_ = []
            for i_img in img:
                img_.append(self.convert(i_img, beta=beta))
            return img_
        else:
            return img

    def contrast(self, img):
        """Contrast distortion."""
        if random.randint(2):
            alpha = random.uniform(self.contrast_lower, self.contrast_upper)
            img_ = []
            for i_img in img:
                img_.append(self.convert(i_img, alpha=alpha))
            return img_
        else:
            return img

    def saturation(self, img):
        """Saturation distortion."""
        if random.randint(2):
            alpha = random.uniform(self.saturation_lower,
                                   self.saturation_upper)
            img_ = []
            for i_img in img:

                i_img = mmcv.bgr2hsv(i_img)
                i_img[:, :, 1] = self.convert(i_img[:, :, 1], alpha=alpha)

                i_img = mmcv.hsv2bgr(i_img)
                img_.append(i_img)
            return img_
        else:
            return img

    def hue(self, img):
        """Hue distortion."""
        if random.randint(2):
            hue_val = random.randint(-self.hue_delta, self.hue_delta)
            img_ = []
            for i_img in img:

                i_img = mmcv.bgr2hsv(i_img)
                i_img[:, :, 0] = (i_img[:, :, 0].astype(int) + hue_val) % 180
                i_img = mmcv.hsv2bgr(i_img)
                img_.append(i_img)
            return img_
        else:
            return img

    def transform(self, results):
        """Call function to perform photometric distortion on images.

        Args:
            results (dict): Result dict from loading pipeline.
        Returns:
            dict: Result dict with images distorted.
        """

        img = []
        for k in img_keys:
            img.append(results[k])
        # random brightness
        img = self.brightness(img)

        # mode == 0 --> do random contrast first
        # mode == 1 --> do random contrast last
        mode = random.randint(2)
        if mode == 1:
            img = self.contrast(img)

        # random saturation
        img = self.saturation(img)

        # random hue
        img = self.hue(img)

        # random contrast
        if mode == 0:
            img = self.contrast(img)

        for i, k in enumerate(img_keys):
            results[k] = img[i]

        return results

    def __repr__(self):
        repr_str = self.__class__.__name__
        repr_str += (f'(brightness_delta={self.brightness_delta}, '
                     f'contrast_range=({self.contrast_lower}, '
                     f'{self.contrast_upper}), '
                     f'saturation_range=({self.saturation_lower}, '
                     f'{self.saturation_upper}), '
                     f'hue_delta={self.hue_delta})')
        return repr_str


@TRANSFORMS.register_module()
class GaussianNoise(BaseTransform):
    """Add Gaussian Noise to images.

    Add Gaussian Noise, with mean 0 and std sigma uniformly sampled from
    sigma_range, to images. And then clamp the images to clamp_range.

    Args:
        sigma_range(list(float) | tuple(float)): Uniformly sample sigma of
            gaussian noise in sigma_range. Default: (0, 0.04)
        clamp_range(list(float) | tuple(float)): The min and max value to clamp
            the images after adding gaussian noise.
            Default: (float('-inf'), float('inf')).
    """

    def __init__(self,
                 sigma_range=(0, 0.04),
                 clamp_range=(float('-inf'), float('inf'))):

        assert isinstance(sigma_range, (list, tuple))
        assert len(sigma_range) == 2
        assert 0 <= sigma_range[0] < sigma_range[1]

        self.sigma_range = sigma_range

        assert isinstance(clamp_range, (list, tuple))
        assert len(clamp_range) == 2
        assert clamp_range[0] < clamp_range[1]

        self.clamp_range = clamp_range

    def transform(self, results):
        """Call function to add gaussian noise to images. And then clamp the
        images to clamp_range.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            dict: Processed results.
        """
        # create new meta 'sigma'
        results['sigma'] = random.uniform(*self.sigma_range)

        for k in img_keys:

            assert results[k].dtype == np.float32, \
                'Before add Gaussian noise, it needs do normalize.'

            results[k] += np.random.randn(
                *results['img_shape']) * results['sigma']

            results[k] = np.clip(
                results[k],
                a_min=self.clamp_range[0],
                a_max=self.clamp_range[1]).astype(np.float32)

        return results

    def __repr__(self):
        return self.__class__.__name__ + f'(sigma_range={self.sigma_range})'


@TRANSFORMS.register_module()
class RandomGamma(BaseTransform):
    """Random gamma correction of images.

    Required Keys:

    - img1
    - img2

    Modified Keys:

    - img1
    - img2

    Added Keys:

    - gamma

    Note: gamma larger than 1 make the shadows darker, while gamma smaller than
    1 make dark regions lighter.

    Args:
        gamma_range(list | tuple): A list or tuple of length 2. Uniformly
            sample gamma from gamma_range. Defaults to (0.7, 1.5).
    """

    def __init__(self, gamma_range: Sequence = (0.7, 1.5)):

        assert isinstance(gamma_range, (list, tuple))

        assert len(gamma_range) == 2

        assert 0 <= gamma_range[0] <= gamma_range[1]

        self.gamma_range = gamma_range

    @cache_randomness
    def _random_gamma(self):
        return random.uniform(*self.gamma_range)

    def transform(self, results: dict) -> dict:
        """Call function to process images using gamma correction.

        Args:
            results (dict): Result dict from loading pipeline.

        Returns:
            dict: Processed results.
        """

        # create new meta 'gamma'
        results['gamma'] = self._random_gamma()

        for k in img_keys:
            results[k] = adjust_gamma(results[k], results['gamma'])

        return results

    def __repr__(self) -> str:
        return self.__class__.__name__ + f'(gamma_range={self.gamma_range})'



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Sequence

import numpy as np
from mmengine.evaluator import BaseMetric

from mmflow.registry import METRICS
from .utils import end_point_error_map


@METRICS.register_module()
class FlowOutliers(BaseMetric):
    """Optical flow outliers metric.

    Args:
        collect_device (str): Device name used for collecting results from
            different ranks during distributed training. Must be 'cpu' or
            'gpu'. Defaults to 'cpu'.
        prefix (str, optional): The prefix that will be added in the metric
            names to disambiguate homonymous metrics of different evaluators.
            If prefix is not provided in the argument, self.default_prefix
            will be used instead. Default: None
    """

    def __init__(self,
                 collect_device: str = 'cpu',
                 prefix: Optional[str] = None) -> None:
        super().__init__(collect_device, prefix)

    def process(self, data_batch: Sequence[dict],
                data_samples: Sequence[dict]) -> None:
        """Process one batch of data samples and predictions. The processed
        results should be stored in ``self.results``, which will be used to
        compute the metrics when all batches have been processed.

        Args:
            data_batch (Sequence[dict]): A batch of data from the dataloader.
            data_samples (Sequence[dict]): A batch of outputs from
                the model.
        """
        gt_flow_list = []
        pred_flow_list = []
        gt_valid_list = []

        for data_sample in data_samples:
            # sintel and kitti only support forward optical flow
            # so here we only evaluate the predicted forward flow
            # tensor with shape (2, H, W) to ndarray with shape (H, W, 2)
            gt_flow = data_sample['gt_flow_fw']['data'].permute(
                1, 2, 0).cpu().numpy()
            pred_flow = data_sample['pred_flow_fw']['data'].permute(
                1, 2, 0).cpu().numpy()

            if data_sample.get('gt_valid_fw', None) is not None:
                # tensor with shape (1, H, W) to ndarray with shape (H, W)
                gt_valid = np.squeeze(
                    data_sample['gt_valid_fw']['data'].cpu().numpy().squeeze())
            else:
                gt_valid = np.ones_like(gt_flow[..., 0])
            gt_flow_list.append(gt_flow)
            pred_flow_list.append(pred_flow)
            gt_valid_list.append(gt_valid)
        fl = self.optical_flow_outliers(pred_flow_list, gt_flow_list,
                                        gt_valid_list)
        self.results.append(fl)

    @staticmethod
    def optical_flow_outliers(flow_pred: Sequence[np.ndarray],
                              flow_gt: Sequence[np.ndarray],
                              valid_gt: Sequence[np.ndarray]) -> float:
        """Calculate percentage of optical flow outliers for KITTI dataset.

        Args:
            flow_pred (list): output list of flow map from flow_estimator
                shape(H, W, 2).
            flow_gt (list): ground truth list of flow map shape(H, W, 2).
            valid_gt (list): the list of valid mask for ground truth with the
                shape (H, W).

        Returns:
            float: optical flow outliers for output.
        """
        out_list = []
        assert len(flow_pred) == len(flow_gt) == len(valid_gt)
        for _flow_pred, _flow_gt, _valid_gt in zip(flow_pred, flow_gt,
                                                   valid_gt):
            epe_map = end_point_error_map(_flow_pred, _flow_gt)
            epe = epe_map.reshape(-1)
            mag_map = np.sqrt(np.sum(_flow_gt**2, axis=-1))
            mag = mag_map.reshape(-1) + 1e-6
            val = _valid_gt.reshape(-1) >= 0.5
            # 3.0 and 0.05 is token from KITTI devkit
            # Inliers are defined as EPE < 3 pixels or < 5%
            out = ((epe > 3.0) & ((epe / mag) > 0.05)).astype(float)
            out_list.append(out[val])
        out_list = np.concatenate(out_list)
        fl = 100 * np.mean(out_list)

        return fl

    def compute_metrics(self, results: list) -> Dict[str, float]:
        """Compute the metrics from processed results.

        Args:
            results (list): The processed results of each batch.

        Returns:
            Dict[str, float]: The computed metrics. The keys are the names of
            the metrics, and the values are corresponding results.
        """
        eval_result = dict(Fl=np.array(results).mean().item())
        return eval_result



# Copyright (c) OpenMMLab. All rights reserved.
from collections.abc import Sequence
from typing import Union

import numpy as np
import torch
from mmcv.transforms import BaseTransform
from mmengine.structures import PixelData
from mmengine.utils import is_str

from mmflow.registry import TRANSFORMS
from mmflow.structures import FlowDataSample


def to_tensor(
    data: Union[np.ndarray, torch.Tensor, Sequence, int,
                float]) -> torch.Tensor:
    """Convert objects of various python types to :obj:`torch.Tensor`.

    Supported types are: :class:`numpy.ndarray`, :class:`torch.Tensor`,
    :class:`Sequence`, :class:`int` and :class:`float`.

    Args:
        data (torch.Tensor | numpy.ndarray | Sequence | int | float): Data to
            be converted.
    """

    if isinstance(data, torch.Tensor):
        return data
    elif isinstance(data, np.ndarray):
        return torch.from_numpy(data)
    elif isinstance(data, Sequence) and not is_str(data):
        return torch.tensor(data)
    elif isinstance(data, int):
        return torch.LongTensor([data])
    elif isinstance(data, float):
        return torch.FloatTensor([data])
    else:
        raise TypeError(f'type {type(data)} cannot be converted to tensor.')


@TRANSFORMS.register_module()
class PackFlowInputs(BaseTransform):
    """Pack the inputs data for the optical flow model.

     The ``img_meta`` item is always populated.  The contents of the
    ``img_meta`` dictionary depends on ``meta_keys``. By default this includes:

        - ``img1_path``: path to the first image file

        - ``img2_path``: path to the first image file

        - ``ori_shape``: original shape of the image as a tuple (h, w, c)

        - ``img_shape``: shape of the image input to the network as a tuple \
            (h, w, c).  Note that images may be zero padded on the \
            bottom/right if the batch tensor is larger than this shape.

        - ``scale_factor``: a float indicating the preprocessing scale

        - ``flip``: a boolean indicating if image flip transform was used

        - ``flip_direction``: the flipping direction

    Args:
        meta_keys (Sequence[str], optional): Meta keys to be collected in
        ``data[img_metas]``. Default: ``('img1_path', 'img2_path',
        'ori_shape', 'img_shape', 'scale_factor', 'flip', 'flip_direction')``

    Args:
        BaseTransform (_type_): _description_

    Returns:
        _type_: _description_
    """
    data_keys = ('gt_flow_fw', 'gt_flow_bw', 'gt_occ_fw', 'gt_occ_bw',
                 'gt_valid_fw', 'gt_valid_bw')

    def __init__(
        self,
        meta_keys=('img1_path', 'img2_path', 'ori_shape', 'img_shape',
                   'scale_factor', 'pad', 'pad_shape', 'flip',
                   'flip_direction')
    ) -> None:

        self.meta_keys = meta_keys

    def transform(self, results: dict) -> dict:
        """Method to pack the input data.

        Args:
            results (dict): Result dict from the data pipeline.

        Returns:
            dict:
            - 'img1' (obj:`torch.Tensor`): The first image data for models.
            - 'img2' (obj:`torch.Tensor`): The second image data for models.
            - 'data_sample' (obj:`FlowDataSample`): The annotation info of the
                sample.
        """
        packed_results = dict()
        data_sample = FlowDataSample()
        if 'img1' in results:
            img1 = results['img1']
            img1 = np.expand_dims(img1, -1) if len(img1.shape) < 3 else img1
            img1 = to_tensor(np.ascontiguousarray(img1.transpose(2, 0, 1)))
        if 'img2' in results:
            img2 = results['img2']
            img2 = np.expand_dims(img2, -1) if len(img2.shape) < 3 else img2
            img2 = to_tensor(np.ascontiguousarray(img2.transpose(2, 0, 1)))
        # inputs shape 2,3,H,W for image1 and image2
        packed_results['inputs'] = torch.stack((img1, img2), dim=0)

        for key in self.data_keys:
            if results.get(key, None) is not None:
                ann_data = results[key]
                if len(ann_data.shape) < 3:
                    ann_data = to_tensor(ann_data[None, ...])
                else:
                    ann_data = to_tensor(
                        np.ascontiguousarray(ann_data.transpose(2, 0, 1)))
                data = PixelData(**dict(data=ann_data))
                data_sample.set_data({key: data})

        img_meta = dict()
        for key in self.meta_keys:
            if key in results:
                img_meta[key] = results[key]
        data_sample.set_metainfo(img_meta)
        packed_results['data_samples'] = data_sample
        return packed_results

    def __repr__(self) -> str:
        repr_str = self.__class__.__name__
        repr_str += f'(meta_keys={self.meta_keys})'
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
from mmengine.hooks import Hook
from mmengine.model import is_model_wrapper
from mmengine.runner import Runner, load_state_dict

from mmflow.registry import HOOKS


@HOOKS.register_module()
class LiteFlowNetStageLoadHook(Hook):
    """Stage loading hook for LiteFlowNet.

    This hook works for loading weights at the previous stage to the additional
    stage in this training.

    Args:
        src_level (str): The source level to be loaded.
        dst_level (str): The level that will load the weights.
    """

    def __init__(self, src_level: str, dst_level: str) -> None:
        super().__init__()

        self.src_level = src_level
        self.dst_level = dst_level

    def before_run(self, runner: Runner) -> None:
        """Before running function of Hook.

        Args:
            runner (Runner): The runner for this training.
        """
        runner.logger.info(
            f'Submodule of LiteFlowNet decoder at {self.dst_level} loads ' +
            f'LiteFlowNet\'s decoder at {self.src_level}')
        if is_model_wrapper(runner.model):
            load_state_dict(
                runner.model.module.decoder.decoders[self.dst_level],
                runner.model.module.decoder.decoders[
                    self.src_level].state_dict())
        else:
            load_state_dict(
                runner.model.decoder.decoders[self.dst_level],
                runner.model.decoder.decoders[self.src_level].state_dict())



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from typing import Optional, Sequence, Union

from mmengine.utils import scandir


def get_data_filename(
        data_dirs: Union[Sequence[str], str],
        suffix: Optional[str] = None,
        exclude: Optional[Sequence[str]] = None) -> Sequence[str]:
    """Get file name from data directory.

    Args:
        data_dirs (list, str): the directory of data
        suffix (str, optional): the suffix for data file. Defaults to None.
        exclude (list, optional): list of files will be excluded.

    Returns:
        list: the list of data file.
    """

    if data_dirs is None:
        return None
    data_dirs = data_dirs \
        if isinstance(data_dirs, (list, tuple)) else [data_dirs]

    suffix = '' if suffix is None else suffix

    if exclude is None:
        exclude = []
    else:
        assert isinstance(exclude, (list, tuple))

    files = []
    for data_dir in data_dirs:
        for f in scandir(data_dir, suffix=suffix):
            if f not in exclude:
                files.append(osp.join(data_dir, f))
    files.sort()
    return files


def load_img_info(data_infos: Sequence[dict], img1_filename: Sequence[str],
                  img2_filename: Sequence[str]) -> None:
    """Load information of images.

    Args:
        data_infos (list): data information.
        img1_filename (list): ordered list of abstract file path of img1.
        img2_filename (list): ordered list of abstract file path of img2.
    """

    num_file = len(img1_filename)
    for i in range(num_file):
        data_info = dict(
            img1_path=img1_filename[i], img2_path=img2_filename[i])
        data_infos.append(data_info)


def load_ann_info(data_infos: Sequence[dict], filename: Sequence[str],
                  filename_key: str) -> None:
    """Load information of annotation.

    Args:
        data_infos (list): data information.
        filename (list): ordered list of abstract file path of annotation.
        filename_key (str): the annotation key e.g. 'flow'.
    """
    assert len(filename) == len(data_infos)
    num_files = len(filename)
    for i in range(num_files):
        data_infos[i][filename_key] = filename[i]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import torch
import torch.nn.functional as F


def multi_level_flow_loss(loss_function,
                          preds_dict: Dict[str, Union[torch.Tensor,
                                                      List[torch.Tensor]]],
                          target: torch.Tensor,
                          weights: Dict[str, float] = dict(
                              level6=0.32,
                              level5=0.08,
                              level4=0.02,
                              level3=0.01,
                              level2=0.005),
                          valid: Optional[torch.Tensor] = None,
                          flow_div: float = 20.,
                          max_flow: float = float('inf'),
                          resize_flow: str = 'downsample',
                          reduction: str = 'sum',
                          scale_as_level: bool = False,
                          **kwargs) -> torch.Tensor:
    """Multi-level endpoint error loss function.

    Args:
        loss_function: pixel-wise loss function for optical flow map.
        preds_dict (dict): multi-level output of predicted optical flow, and
            the contain in dict is a Tensor or list of Tensor with shape
            (B, 1, H_l, W_l), where l indicates the level.
        target (Tensor): ground truth of optical flow with shape (B, 2, H, W).
        weights (dict): manual rescaling weights given to the loss of flow map
            at each level, and the keys in weights must correspond to predicted
            dict. Defaults to dict(
            level6=0.32, level5=0.08, level4=0.02, level3=0.01, level2=0.005).
        valid (Tensor, optional): valid mask for optical flow.
            Defaults to None.
        flow_div (float): the divisor used to scale down ground truth.
            Defaults to 20.
        max_flow (float): maximum value of optical flow, if some pixel's flow
            of target is larger than it, this pixel is not valid. Default to
            float('inf').
        reduction (str): the reduction to apply to the output:'none', 'mean',
            'sum'. 'none': no reduction will be applied and will return a
            full-size epe map, 'mean': the mean of the epe map is taken, 'sum':
            the epe map will be summed but averaged by batch_size.
            Default: 'sum'.
        resize_flow (str): mode for reszing flow: 'downsample' and 'upsample',
            as multi-level predicted outputs don't match the ground truth.
            If set to 'downsample', it will downsample the ground truth, and
            if set to 'upsample' it will upsample the predicted flow, and
            'upsample' is used for sparse flow map as no generic interpolation
            mode can resize a ground truth of sparse flow correctly.
            Default to 'downsample'.
        scale_as_level (bool): Whether flow for each level is at its native
            spatial resolution. If `'scale_as_level'` is True, the ground
            truth is scaled at different levels, if it is False, the ground
            truth will not be scaled. Default to False.
        kwargs: arguments for loss_function.

    Returns:
        Tensor: end-point error loss.
    """

    assert isinstance(weights, dict)

    assert list(preds_dict.keys()).sort() == list(weights.keys()).sort(), \
        'Error: Keys of prediction do not match keys of weights!'

    mag = torch.sum(target**2, dim=1).sqrt()

    if valid is None:
        valid = torch.ones_like(target[:, 0, :, :])
    else:
        valid = ((valid >= 0.5) & (mag < max_flow)).to(target)

    target_div = target / flow_div

    c_org, h_org, w_org = target.shape[1:]
    assert c_org == 2, f'The channels ground truth must be 2, but got {c_org}'

    loss = 0

    for level in weights.keys():

        # predict more than one flow map at one level
        cur_pred = preds_dict[level] if isinstance(
            preds_dict[level], (tuple, list)) else [preds_dict[level]]

        num_preds = len(cur_pred)

        b, _, h, w = cur_pred[0].shape

        scale_factor = torch.Tensor([
            float(w / w_org), float(h / h_org)
        ]).to(target) if scale_as_level else torch.Tensor([1., 1.]).to(target)

        cur_weight = weights.get(level)

        if resize_flow == 'downsample':
            # down sample ground truth following irr solution
            # https://github.com/visinf/irr/blob/master/losses.py#L16
            cur_target = F.adaptive_avg_pool2d(target_div, [h, w])
            cur_valid = F.adaptive_max_pool2d(valid, [h, w])
        else:
            cur_target = target_div
            cur_valid = valid

        loss_map = torch.zeros_like(cur_target[:, 0, ...])

        for i_pred in cur_pred:

            if resize_flow == 'upsample':
                # up sample predicted flow following pwcnet and irr solution
                # https://github.com/visinf/irr/blob/master/losses.py#L20
                # when training sparse flow dataset, as no generic
                # interpolation mode can resize a ground truth of sparse flow
                # correctly.
                i_pred = F.interpolate(
                    i_pred,
                    size=cur_target.shape[2:],
                    mode='bilinear',
                    align_corners=False)

            cur_target = torch.einsum('b c h w, c -> b c h w', cur_target,
                                      scale_factor)

            loss_map += loss_function(i_pred, cur_target, **kwargs) * cur_valid

            if reduction == 'mean':
                loss += loss_map.sum() / (cur_valid.sum() + 1e-8) * cur_weight
            elif reduction == 'sum':
                loss += loss_map.sum() / b * cur_weight

    return loss / num_preds



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import json
import os.path as osp
from glob import glob

from mmengine.utils import mkdir_or_exist
from utils import get_data_filename


def parse_args():
    parser = argparse.ArgumentParser(
        description='FlyingThings3D dataset preparation')
    parser.add_argument(
        '--data-root',
        type=str,
        default='data/flyingthings3d',
        help='Directory for dataset.')
    parser.add_argument(
        '--save-dir',
        type=str,
        default='data/flyingthings3d/',
        help='Directory to save the annotation files for '
        'FlyingThings3D dataset')
    args = parser.parse_args()

    return args


def main():
    args = parse_args()
    flow_fw_dir = 'into_future'
    flow_bw_dir = 'into_past'
    flow_suffix = '.pfm'
    img_suffix = '.png'
    scene = ['left', 'right']

    def _get_data_list(subset_dir):
        imgs_dirs = []
        flow_fw_dirs = []
        flow_bw_dirs = []
        for i_scene in scene:
            tmp_imgs_dirs = glob(
                osp.join(args.data_root, 'frames_cleanpass',
                         subset_dir + '/*/*'))
            tmp_imgs_dirs = [osp.join(f, i_scene) for f in tmp_imgs_dirs]
            imgs_dirs += tmp_imgs_dirs
            tmp_flow_fw_dirs = glob(
                osp.join(args.data_root, 'optical_flow',
                         subset_dir + '/*/*/' + flow_fw_dir))

            tmp_flow_fw_dirs = [osp.join(f, i_scene) for f in tmp_flow_fw_dirs]
            flow_fw_dirs += tmp_flow_fw_dirs

            tmp_flow_bw_dirs = glob(
                osp.join(args.data_root, 'optical_flow',
                         subset_dir + '/*/*/' + flow_bw_dir))
            tmp_flow_bw_dirs = [osp.join(f, i_scene) for f in tmp_flow_bw_dirs]
            flow_bw_dirs += tmp_flow_bw_dirs
        img1_filenames = []
        img2_filenames = []
        flow_fw_filenames = []
        flow_bw_filenames = []
        for idir, fw_dir, bw_dir in zip(imgs_dirs, flow_fw_dirs, flow_bw_dirs):

            img_filenames_ = get_data_filename(idir, img_suffix)
            img_filenames_.sort()
            flow_fw_filenames_ = get_data_filename(fw_dir, flow_suffix)
            flow_fw_filenames_.sort()
            flow_bw_filenames_ = get_data_filename(bw_dir, flow_suffix)
            flow_bw_filenames_.sort()

            img1_filenames.extend(img_filenames_[:-1])
            img2_filenames.extend(img_filenames_[1:])
            flow_fw_filenames.extend(flow_fw_filenames_[:-1])
            flow_bw_filenames.extend(flow_bw_filenames_[1:])

        data_list = []

        assert len(img1_filenames) == len(img2_filenames) == len(
            flow_fw_filenames) == len(flow_bw_filenames)

        for i_img1, i_img2, i_flow_fw, i_flow_bw in zip(
                img1_filenames, img2_filenames, flow_fw_filenames,
                flow_bw_filenames):

            i_scene = scene[0] if scene[0] in i_img1 else scene[1]

            data_info = dict(
                scene=i_scene,
                pass_style='clean',
                img1_path=i_img1,
                img2_path=i_img2,
                flow_fw_path=i_flow_fw,
                flow_bw_path=i_flow_bw,
            )
            data_info_final = dict(
                scene=i_scene,
                pass_style='final',
                img1_path=i_img1.replace('clean', 'final'),
                img2_path=i_img2.replace('clean', 'final'),
                flow_fw_path=i_flow_fw,
                flow_bw_path=i_flow_bw,
            )

            data_list.append(data_info)
            data_list.append(data_info_final)

        mkdir_or_exist(args.save_dir)
        if subset_dir == 'TRAIN':
            annotation_file = osp.join(args.save_dir, 'train.json')
        else:
            annotation_file = osp.join(args.save_dir, 'test.json')
        with open(annotation_file, 'w') as jsonfile:
            json.dump({'data_list': data_list, 'metainfo': {}}, jsonfile)

    _get_data_list('TRAIN')
    _get_data_list('TEST')


if __name__ == '__main__':
    main()



img_norm_cfg = dict(mean=[0., 0., 0.], std=[255., 255., 255.], to_rgb=False)

crop_size = (320, 768)

sintel_global_transform = dict(
    translates=(0.05, 0.05),
    zoom=(1.0, 1.2),
    shear=(0.95, 1.1),
    rotate=(-5., 5.))

sintel_relative_transform = dict(
    translates=(0.00375, 0.00375),
    zoom=(0.985, 1.015),
    shear=(1.0, 1.0),
    rotate=(-1.0, 1.0))

sintel_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', with_occ=True),
    dict(
        type='ColorJitter',
        brightness=0.5,
        contrast=0.5,
        saturation=0.5,
        hue=0.5),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='GaussianNoise', sigma_range=(0, 0.04), clamp_range=(0., 1.)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=sintel_global_transform,
        relative_transform=sintel_relative_transform),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='DefaultFormatBundle'),
    dict(
        type='Collect',
        keys=['imgs', 'flow_gt'],
        meta_keys=[
            'img_fields', 'ann_fields', 'filename1', 'filename2',
            'ori_filename1', 'ori_filename2', 'filename_flow',
            'ori_filename_flow', 'ori_shape', 'img_shape', 'img_norm_cfg'
        ]),
]

sintel_test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='InputPad', exponent=6),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='TestFormatBundle'),
    dict(
        type='Collect',
        keys=['imgs'],
        meta_keys=[
            'flow_gt', 'filename1', 'filename2', 'ori_filename1',
            'ori_filename2', 'ori_shape', 'img_shape', 'img_norm_cfg',
            'scale_factor', 'pad_shape', 'pad'
        ])
]

sintel_clean_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    scene=[
        'alley_1', 'alley_2', 'ambush_4', 'ambush_5', 'ambush_7', 'bamboo_1',
        'bandage_1', 'bandage_2', 'cave_2', 'market_2', 'market_5',
        'mountain_1', 'shaman_2', 'shaman_3', 'sleeping_1', 'sleeping_2',
        'temple_3'
    ],
    pass_style='clean')

sintel_final_train = dict(
    type='Sintel',
    pipeline=sintel_train_pipeline,
    data_root='data/Sintel',
    test_mode=False,
    scene=[
        'alley_1', 'alley_2', 'ambush_4', 'ambush_5', 'ambush_7', 'bamboo_1',
        'bandage_1', 'bandage_2', 'cave_2', 'market_2', 'market_5',
        'mountain_1', 'shaman_2', 'shaman_3', 'sleeping_1', 'sleeping_2',
        'temple_3'
    ],
    pass_style='final')

sintel_clean_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=True,
    scene=[
        'ambush_2', 'ambush_6', 'bamboo_2', 'cave_4', 'market_6', 'temple_2'
    ],
    pass_style='clean')

sintel_final_test = dict(
    type='Sintel',
    pipeline=sintel_test_pipeline,
    data_root='data/Sintel',
    test_mode=True,
    scene=[
        'ambush_2', 'ambush_6', 'bamboo_2', 'cave_4', 'market_6', 'temple_2'
    ],
    pass_style='final')

kitti_global_transform = dict(
    translates=(0.02, 0.02),
    zoom=(0.98, 1.02),
    shear=(1.0, 1.0),
    rotate=(-0.5, 0.5))

kitti_relative_transform = dict(
    translates=(0.0025, 0.0025),
    zoom=(0.99, 1.01),
    shear=(1.0, 1.0),
    rotate=(-0.5, 0.5))

sparse_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', sparse=True),
    dict(type='RandomCrop', crop_size=crop_size * 2),
    dict(
        type='ColorJitter',
        brightness=0.05,
        contrast=0.2,
        saturation=0.25,
        hue=0.1),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='GaussianNoise', sigma_range=(0, 0.02), clamp_range=(0., 1.)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=kitti_global_transform,
        relative_transform=kitti_relative_transform),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='DefaultFormatBundle'),
    dict(
        type='Collect',
        keys=['imgs', 'flow_gt', 'valid'],
        meta_keys=[
            'img_fields', 'ann_fields', 'filename1', 'filename2',
            'ori_filename1', 'ori_filename2', 'filename_flow',
            'ori_filename_flow', 'ori_shape', 'img_shape', 'img_norm_cfg'
        ]),
]

kitti2015_train = dict(
    type='KITTI2015',
    data_root='data/kitti2015',
    pipeline=sparse_train_pipeline,
    test_mode=False)

hd1k_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', sparse=True),
    dict(type='RandomCrop', crop_size=(436, 1024)),
    dict(
        type='ColorJitter',
        brightness=0.05,
        contrast=0.2,
        saturation=0.25,
        hue=0.1),
    dict(type='RandomGamma', gamma_range=(0.7, 1.5)),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='GaussianNoise', sigma_range=(0, 0.02), clamp_range=(0., 1.)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.5, direction='vertical'),
    dict(
        type='RandomAffine',
        global_transform=kitti_global_transform,
        relative_transform=kitti_relative_transform),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='DefaultFormatBundle'),
    dict(
        type='Collect',
        keys=['imgs', 'flow_gt', 'valid'],
        meta_keys=[
            'img_fields', 'ann_fields', 'filename1', 'filename2',
            'ori_filename1', 'ori_filename2', 'filename_flow',
            'ori_filename_flow', 'ori_shape', 'img_shape', 'img_norm_cfg'
        ]),
]

hd1k_train = dict(
    type='HD1K',
    data_root='data/hd1k',
    pipeline=hd1k_train_pipeline,
    test_mode=False),

data = dict(
    train_dataloader=dict(
        samples_per_gpu=1,
        workers_per_gpu=2,
        drop_last=True,
        sample_ratio=(0.5, 0.25, 0.25),
        persistent_workers=True),
    val_dataloader=dict(
        samples_per_gpu=1,
        workers_per_gpu=5,
        shuffle=False,
        persistent_workers=True),
    test_dataloader=dict(samples_per_gpu=1, workers_per_gpu=5, shuffle=False),
    train=[[sintel_clean_train, sintel_final_train], kitti2015_train,
           hd1k_train],
    val=dict(
        type='ConcatDataset',
        datasets=[sintel_clean_test, sintel_final_test],
        separate_eval=True),
    test=dict(
        type='ConcatDataset',
        datasets=[sintel_clean_test, sintel_final_test],
        separate_eval=True))



# Copyright (c) OpenMMLab. All rights reserved.
import os.path as osp
from glob import glob
from typing import Callable, List, Optional, Sequence, Union

from mmengine.dataset import BaseDataset as MME_BaseDataset

from mmflow.registry import DATASETS
from .utils import load_ann_info, load_img_info


@DATASETS.register_module()
class HD1K(MME_BaseDataset):
    """HD1K dataset.

    Args:
        ann_file (str): Annotation file path. Defaults to ''.
        metainfo (dict, optional): Meta information for dataset, such as class
            information. Defaults to None.
        data_root (str, optional): The root directory for ``data_prefix`` and
            ``ann_file``. Defaults to None.
        data_prefix (dict, optional): Prefix for training data. Defaults to
            dict(img=None, ann=None).
        filter_cfg (dict, optional): Config for filter data. Defaults to None.
        indices (int or Sequence[int], optional): Support using first few
            data in annotation file to facilitate training/testing on a smaller
            dataset. Defaults to None which means using all ``data_infos``.
        serialize_data (bool, optional): Whether to hold memory using
            serialized objects, when enabled, data loader workers can use
            shared RAM from master process instead of making a copy. Defaults
            to True.
        pipeline (list, optional): Processing pipeline. Defaults to [].
        test_mode (bool, optional): ``test_mode=True`` means in test phase.
            Defaults to False.
        lazy_init (bool, optional): Whether to load annotation during
            instantiation. In some cases, such as visualization, only the meta
            information of the dataset is needed, which is not necessary to
            load annotation file. ``Basedataset`` can skip load annotations to
            save time by set ``lazy_init=False``. Defaults to False.
        max_refetch (int, optional): If ``Basedataset.prepare_data`` get a
            None img. The maximum extra number of cycles to get a valid
            image. Defaults to 1000.
    """
    METAINFO = dict()

    def __init__(self,
                 ann_file: str = '',
                 metainfo: Optional[dict] = None,
                 data_root: Optional[str] = None,
                 data_prefix: dict = dict(),
                 filter_cfg: Optional[dict] = None,
                 indices: Optional[Union[int, Sequence[int]]] = None,
                 serialize_data: bool = True,
                 pipeline: List[Union[dict, Callable]] = [],
                 test_mode: bool = False,
                 lazy_init: bool = False,
                 max_refetch: int = 1000) -> None:
        super().__init__(ann_file, metainfo, data_root, data_prefix,
                         filter_cfg, indices, serialize_data, pipeline,
                         test_mode, lazy_init, max_refetch)

    def load_data_list(self) -> List[dict]:
        """Load ``data_list``

        ``data_list`` can be load from an annotation file named as
        ``self.ann_file`` or by parsing dataset path.

        Returns:
            list[dict]: A list of annotation.
        """
        if self.ann_file.endswith('json'):
            # load data_list with annotation file
            return super().load_data_list()
        else:
            # load data_list by path parsing
            self.load_data_info()
            return self.data_list

    def load_data_info(self) -> None:
        """Load data information, including file path of image1, image2 and
        optical flow."""

        self.img1_dir = osp.join(self.data_root, 'hd1k_input/image_2')
        self.img2_dir = osp.join(self.data_root, 'hd1k_input/image_2')
        self.flow_dir = osp.join(self.data_root, 'hd1k_flow_gt/flow_occ')

        img1_filenames = []
        img2_filenames = []
        flow_filenames = []

        seq_ix = 0
        while True:
            flows = sorted(
                glob(osp.join(self.flow_dir, '%06d_*.png' % seq_ix)))
            images = sorted(
                glob(osp.join(self.img1_dir, '%06d_*.png' % seq_ix)))

            if len(flows) == 0:
                break

            for i in range(len(flows) - 1):
                flow_filenames += [flows[i]]
                img1_filenames += [images[i]]
                img2_filenames += [images[i + 1]]

            seq_ix += 1

        load_img_info(self.data_list, img1_filenames, img2_filenames)
        load_ann_info(self.data_list, flow_filenames, 'flow_fw_path')



train_cfg = dict(by_epoch=False, max_iters=100000, val_interval=10000)
val_cfg = dict(type='ValLoop')
test_cfg = dict(type='TestLoop')

# optimizer
optim_wrapper = dict(
    type='OptimWrapper',
    optimizer=dict(
        type='AdamW',
        lr=0.0004,
        betas=(0.9, 0.999),
        eps=1e-08,
        weight_decay=0.0001,
        amsgrad=False),
    clip_grad=dict(max_norm=1.))

# learning policy
param_scheduler = dict(
    type='OneCycleLR',
    eta_max=0.0004,
    total_steps=100100,
    pct_start=0.05,
    anneal_strategy='linear',
    by_epoch=False)

default_hooks = dict(
    timer=dict(type='IterTimerHook'),
    logger=dict(type='LoggerHook', interval=50, log_metric_by_epoch=False),
    param_scheduler=dict(type='ParamSchedulerHook'),
    checkpoint=dict(type='CheckpointHook', interval=10000, by_epoch=False),
    sampler_seed=dict(type='DistSamplerSeedHook'),
    visualization=dict(type='FlowVisualizationHook'))



crop_size = (288, 960)

# KITTI config
kitti_train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', sparse=True),
    dict(
        type='ColorJitter',
        asymmetric_prob=0.0,
        brightness=0.4,
        contrast=0.4,
        saturation=0.4,
        hue=0.5 / 3.14),
    dict(type='Erase', prob=0.5, bounds=[50, 100], max_num=3),
    dict(
        type='SpacialTransform',
        spacial_prob=0.8,
        stretch_prob=0.8,
        crop_size=crop_size,
        min_scale=-0.2,
        max_scale=0.4,
        max_stretch=0.2),
    dict(type='RandomCrop', crop_size=crop_size),
    dict(type='PackFlowInputs')
]
kitti_train = dict(
    type='KITTI2015',
    data_root='data/kitti2015',
    pipeline=kitti_train_pipeline,
    test_mode=False)

kitti_test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations', sparse=True),
    dict(type='InputPad', exponent=3),
    dict(type='PackFlowInputs')
]

kitti2015_val_test = dict(
    type='KITTI2015',
    data_root='data/kitti2015',
    pipeline=kitti_test_pipeline,
    test_mode=False)

train_dataloader = dict(
    batch_size=2,
    num_workers=5,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    drop_last=True,
    persistent_workers=True,
    dataset=kitti_train)
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    sampler=dict(type='DefaultSampler', shuffle=False),
    drop_last=False,
    persistent_workers=True,
    dataset=kitti2015_val_test)
test_dataloader = val_dataloader

val_evaluator = [
    dict(type='EndPointError', prefix='KITTI2015'),
    dict(type='FlowOutliers', prefix='KITTI2015')
]
test_evaluator = val_evaluator



# Copyright (c) OpenMMLab. All rights reserved.
from abc import abstractmethod
from typing import Dict, List, Optional, Sequence, Union

import torch.nn.functional as F
from mmengine.model import BaseModule
from mmengine.structures import PixelData
from torch import Tensor

from mmflow.structures import FlowDataSample
from mmflow.utils import OptSampleList, SampleList, TensorDict


class BaseDecoder(BaseModule):
    """Base class for decoder.

    Args:
        init_cfg (dict, list, optional): Config dict of weights initialization.
            Default: None.
    """

    def __init__(self, init_cfg: Optional[Union[dict, list]] = None) -> None:

        super().__init__(init_cfg=init_cfg)

    @abstractmethod
    def forward(self, *args, **kwargs):
        """Placeholder of forward function."""
        pass

    @abstractmethod
    def loss(self, *args, **kwargs):
        """Placeholder of forward function when model training."""
        pass

    @abstractmethod
    def predict(self, *args, **kwargs):
        """Placeholder of forward function when model testing."""
        pass

    @abstractmethod
    def loss_by_feat(self, *args, **kwargs) -> TensorDict:
        """Placeholder for model computing losses."""
        pass

    def postprocess_result(
            self,
            flow_results: List[Dict],
            data_samples: OptSampleList = None) -> Sequence[FlowDataSample]:
        """Reverted flow as original size of ground truth.

        Args:
            flow_results (List[Dict]): List of predicted results.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        only_prediction = False
        if data_samples is None:
            data_samples = []
            only_prediction = True
        else:
            assert len(flow_results) == len(data_samples)

        for i in range(len(flow_results)):
            if only_prediction:
                prediction = FlowDataSample()
                for key, f in flow_results[i].items():
                    prediction.set_data({
                        'pred_' + key:
                        PixelData(**{'data': flow_results[i]})
                    })
                data_samples.append(prediction)
            else:
                img_meta = data_samples[i].metainfo
                ori_H, ori_W = img_meta['ori_shape']
                pad = img_meta.get('pad', None)
                w_scale, h_scale = img_meta.get('scale_factor', (None, None))
                for key, f in flow_results[i].items():
                    if f is not None:
                        # shape is 2, H, W
                        H, W = f.shape[1:]
                        if pad is not None:
                            f = f[:, pad[0][0]:(H - pad[0][1]),
                                  pad[1][0]:(W - pad[1][1])]

                        elif (w_scale is not None and h_scale is not None):
                            f = F.interpolate(
                                f[None],
                                size=(ori_H, ori_W),
                                mode='bilinear',
                                align_corners=False).squeeze(0)
                            f[0, :, :] = f[0, :, :] / w_scale
                            f[1, :, :] = f[1, :, :] / h_scale
                    data_samples[i].set_data(
                        {'pred_' + key: PixelData(**{'data': f})})
            return data_samples

    def predict_by_feat(self,
                        flow_results: Tensor,
                        data_samples: OptSampleList = None) -> SampleList:
        """Predict list of obj:`FlowDataSample` from flow tensor.

        Args:
            flow_results (Tensor): Predicted flow tensor.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """
        if data_samples is None:
            flow_results = flow_results * self.flow_div
            # unravel batch dim,
            flow_results = list(flow_results)
            flow_results = [dict(flow_fw=f) for f in flow_results]
            return self.postprocess_result(flow_results, data_samples=None)

        H, W = data_samples[0].metainfo['img_shape'][:2]
        # resize flow to the size of images after augmentation.
        flow_results = F.interpolate(
            flow_results, size=(H, W), mode='bilinear', align_corners=False)

        flow_results = flow_results * self.flow_div

        # unravel batch dim,
        flow_results = list(flow_results)
        flow_results = [dict(flow_fw=f) for f in flow_results]

        return self.postprocess_result(flow_results, data_samples=data_samples)



# Copyright (c) OpenMMLab. All rights reserved.
import argparse
import os.path as osp
import time

import numpy as np
import torch
from mmengine import Config
from mmengine.fileio import dump
from mmengine.runner import Runner, load_checkpoint
from mmengine.utils import mkdir_or_exist

from mmflow.models import build_flow_estimator
from mmflow.utils import register_all_modules


def parse_args():
    parser = argparse.ArgumentParser(description='MMFlow benchmark a model')
    parser.add_argument('config', help='test config file path')
    parser.add_argument('checkpoint', help='checkpoint file')
    parser.add_argument(
        '--log-interval', type=int, default=50, help='interval of logging')
    parser.add_argument(
        '--work-dir',
        help=('if specified, the results will be dumped '
              'into the directory as json'))
    parser.add_argument(
        '--warm_up',
        type=int,
        default=5,
        help='the warm-up iterations should be skipped')
    parser.add_argument(
        '--total_iters',
        type=int,
        default=200,
        help='the total iterations during benchmark')
    parser.add_argument('--repeat-times', type=int, default=1)
    args = parser.parse_args()
    return args


def main():
    args = parse_args()
    register_all_modules()
    cfg = Config.fromfile(args.config)

    timestamp = time.strftime('%Y%m%d_%H%M%S', time.localtime())
    if args.work_dir is not None:
        mkdir_or_exist(osp.abspath(args.work_dir))
        json_file = osp.join(args.work_dir, f'fps_{timestamp}.json')
    else:
        # use config filename as default work_dir if cfg.work_dir is None
        work_dir = osp.join('./work_dirs',
                            osp.splitext(osp.basename(args.config))[0])
        mkdir_or_exist(osp.abspath(work_dir))
        json_file = osp.join(work_dir, f'fps_{timestamp}.json')

    repeat_times = args.repeat_times
    # set cudnn_benchmark
    torch.backends.cudnn.benchmark = False

    benchmark_dict = dict(config=args.config, unit='img / s')
    overall_fps_list = []
    if isinstance(cfg.test_dataloader, list):
        test_dataloader = cfg.test_dataloader[0]
    else:
        test_dataloader = cfg.test_dataloader
    test_dataloader.batch_size = 1
    for time_index in range(repeat_times):
        print(f'Run {time_index + 1}:')
        # build the dataloader
        data_loader = Runner.build_dataloader(test_dataloader)

        model = build_flow_estimator(cfg.model)

        if 'checkpoint' in args and osp.exists(args.checkpoint):
            load_checkpoint(model, args.checkpoint, map_location='cpu')

        model.eval()

        if torch.cuda.is_available():
            model = model.cuda()

        # the first several iterations may be very slow so skip them
        num_warmup = args.warm_up
        pure_inf_time = 0
        total_iters = args.total_iters

        # benchmark with 200 image and take the average
        for i, data in enumerate(data_loader):
            data = model.data_preprocessor(data, True)

            torch.cuda.synchronize()
            start_time = time.perf_counter()

            with torch.no_grad():
                model(**data, mode='predict')

            torch.cuda.synchronize()
            elapsed = time.perf_counter() - start_time

            if i >= num_warmup:
                pure_inf_time += elapsed
                if (i + 1) % args.log_interval == 0:
                    fps = (i + 1 - num_warmup) / pure_inf_time
                    print(f'Done image [{i + 1:<3}/ {total_iters}], '
                          f'fps: {fps:.2f} img / s')

            if (i + 1) == total_iters:
                fps = (i + 1 - num_warmup) / pure_inf_time
                print(f'Overall fps: {fps:.2f} img / s')
                benchmark_dict[f'overall_fps_{time_index + 1}'] = round(fps, 2)
                overall_fps_list.append(fps)
                break

    benchmark_dict['average_fps'] = round(np.mean(overall_fps_list), 2)
    benchmark_dict['fps_variance'] = round(np.var(overall_fps_list), 4)
    print(f'Average fps of {repeat_times} evaluations: '
          f'{benchmark_dict["average_fps"]}')
    print(f'The variance of {repeat_times} evaluations: '
          f'{benchmark_dict["fps_variance"]}')
    dump(benchmark_dict, json_file, indent=4)


if __name__ == '__main__':
    main()
